import ast
import logging
import io
import os
import platform
import queue
import re
import subprocess
import sys
import textwrap
import threading
import time
import tokenize
import traceback
import webbrowser
from pathlib import Path
from queue import Queue
from textwrap import dedent
from time import sleep
from tkinter import ttk
from tkinter.filedialog import askopenfilename
from tkinter.messagebox import showerror
from typing import Optional

import jedi
import serial.tools.list_ports
from serial import SerialException

from thonny import common, get_runner, get_shell, get_workbench
from thonny.common import (
    BackendEvent,
    InlineResponse,
    MessageFromBackend,
    ToplevelCommand,
    ToplevelResponse,
)
from thonny.config_ui import ConfigurationPage
from thonny.misc_utils import find_volumes_by_name
from thonny.plugins.backend_config_page import BackendDetailsConfigPage
from thonny.running import BackendProxy
from thonny.ui_utils import SubprocessDialog, create_string_var

EOT = b"\x04"
NORMAL_PROMPT = b">>> "
# first prompt when switching to raw mode (or after soft reboot in raw mode)
FIRST_RAW_PROMPT = b"raw REPL; CTRL-B to exit\r\n>"
RAW_PROMPT = b">"
TIMEOUT = 0.1
EOT_WITH_RAW_PROMPT = "\x04>"
THONNY_START = "<ForThonny>"
THONNY_END = "</ForThonny>"
THONNY_MSG_START = b"\x02"
NEWLINE = "\n"

# TODO: Current code has some preparations in place to make automatic initial interrupt optional
# It's not so easy, though (initial interrupt also fetches some required info etc)
_AUTOMATIC_INTERRUPT = True


class MicroPythonProxy(BackendProxy):
    def __init__(self, clean):
        super().__init__(clean)
        self._non_serial_msg_queue = Queue()
        self._last_toplevel_command = None
        self._has_been_idle = False
        self._ctrl_c_notice_has_been_removed = False

        self._baudrate = 115200

        self._reading_cancelled = False
        self._welcome_text = None
        self._discarded_bytes = bytearray()
        self._builtins_info = self._get_builtins_info()
        # TODO: provide default builtins for script completion
        self._builtin_modules = []

        self.__idle = False

        self._serial = self._create_serial()

        if self._serial is not None and (clean or _AUTOMATIC_INTERRUPT):
            try:
                self._interrupt_to_prompt(clean)
                self._builtin_modules = self._fetch_builtin_modules()
            except TimeoutError:
                read_bytes = bytes(self._discarded_bytes + self._serial._read_buffer)
                self._show_error_connect_again(
                    "Could not connect to REPL.\n"
                    + "Make sure your device has suitable firmware and is not in bootloader mode!\n"
                    + "Bytes read: "
                    + str(read_bytes)
                    + "\nDisconnecting."
                )
                self.disconnect()
            except:
                self.disconnect()
                raise

        self._start_time = time.time()

    def send_command(self, cmd):
        if isinstance(cmd, ToplevelCommand):
            self._last_toplevel_command = cmd

        if cmd.name in ["editor_autocomplete", "cd", "dump_api_info", "lsdevice"]:
            # Works even without connection to the board
            return super().send_command(cmd)
        elif self._serial is None:
            return "discard"
        elif self.idle:
            try:
                if not self._serial.buffers_are_empty():
                    discarded = self._serial.read_all()
                    self._send_error_to_shell(
                        "Warning: when issuing %r,\nincoming was not emtpy: %r"
                        % (cmd, discarded)
                    )
                return super().send_command(cmd)
            except SerialException as e:
                self._handle_serial_exception(e)
                return "discard"
        else:
            return "postpone"

    def send_program_input(self, data: str) -> None:

        # TODO: what if there is a previous unused data waiting
        assert self._serial.outgoing_is_empty()

        assert data.endswith("\n")
        if not data.endswith("\r\n"):
            input_str = data[:-1] + "\r\n"

        data = input_str.encode("utf-8")

        try:
            self._serial.write(data)
            # Try to consume the echo

            try:
                echo = self._serial.read(len(data))
            except queue.Empty:
                # leave it.
                logging.warning("Timeout when reading echo")
                return

            if echo != data:
                # because of autoreload? timing problems? interruption?
                # Leave it.
                logging.warning("Unexpected echo. Expected %s, got %s" % (data, echo))
                self._serial.unread(echo)

        except SerialException as e:
            self._handle_serial_exception(e)

    def fetch_next_message(self):
        if not self._non_serial_msg_queue.empty():
            msg = self._non_serial_msg_queue.get_nowait()

        elif self._serial is not None:
            # Provide guidance for Ctrl-C
            if time.time() - self._start_time > 0.5:
                if not self._has_been_idle:
                    """TODO: get_shell().set_notice("Use Ctrl-C to interrupt the program and/or enter the REPL")"""
                else:
                    if not self._ctrl_c_notice_has_been_removed:
                        """TODO: get_shell().set_notice(None)"""
                        self._ctrl_c_notice_has_been_removed = True
                        # TODO: fetch required info if automatic interrupt is disabled

            # get the message
            try:
                msg = self._read_next_serial_message()
            except SerialException as e:
                self._handle_serial_exception(e)
                return None
        else:
            msg = None

        return self.transform_message(msg)

    def interrupt(self):
        if self._serial is None:
            return

        try:
            self.idle = False
            self._serial.reset_output_buffer()
            self._serial.write(b"\x03")

            # Wait a bit to avoid the situation where part of the prompt will
            # be treated as output and whole prompt is not detected.
            # (Happened with Calliope)
            sleep(0.1)
        except SerialException as e:
            self._handle_serial_exception(e)

    def destroy(self):
        self.disconnect()

    def disconnect(self):
        if self._serial is not None:
            try:
                self._serial.close()
                self._send_text_to_shell(
                    "\n\nConnection closed.\nSelect Run → Stop/Restart or press Ctrl+F2 to connect again.",
                    "stdout",
                )
            except Exception as e:
                logging.exception("Problem when closing serial")
                self._send_error_to_shell(
                    "Problem when closing serial connection: " + str(e)
                )

            self._serial = None

    def is_connected(self):
        return self._serial is not None

    def is_functional(self):
        return self.is_connected()

    def _create_serial(self):
        self.port = get_workbench().get_option(self.backend_name + ".port")

        if self.port is None:
            self._send_text_to_shell(
                'Not connected. Choose "Tools → Options → Backend" to change.', "stdout"
            )
            return None

        if self.port == "auto":
            potential = self._detect_potential_ports()
            if len(potential) == 1:
                self.port = potential[0][0]
            else:
                message = dedent(
                    """\
                    Couldn't find the device automatically. 
                    Check the connection (making sure the device is not in bootloader mode)
                    or choose "Tools → Options → Backend" to select the port manually."""
                )

                if len(potential) > 1:
                    _, descriptions = zip(*potential)
                    message += "\n\nLikely candidates are:\n * " + "\n * ".join(
                        descriptions
                    )

                self._show_error_connect_again(message)
                return None

        try:
            return SerialHelper(self.port, baudrate=self._baudrate)
        except SerialException as error:
            traceback.print_exc()
            message = (
                "Unable to connect to " + self.port + "\n" + "Error: " + str(error)
            )

            # TODO: check if these error codes also apply to Linux and Mac
            if error.errno == 13 and platform.system() == "Linux":
                # TODO: check if user already has this group
                message += "\n\n" + dedent(
                    """\
                Try adding yourself to the 'dialout' group:
                > sudo usermod -a -G dialout <username>
                (NB! This needs to be followed by reboot or logging out and logging in again!)"""
                )

            elif "PermissionError" in message:
                message += "\n\n" + dedent(
                    """\
                If you have serial connection to the device from another program,
                then disconnect it there."""
                )

            elif error.errno == 16:
                message += "\n\n" + "Try restarting the device."

            self._show_error_connect_again(message)

            return None

    def _show_error_connect_again(self, msg):
        self._send_error_to_shell(
            msg
            + "\n\nCheck the configuration, select Run → Stop/Restart or press Ctrl+F2 to try again."
            + "\n(On some occasions it helps to wait before trying again.)"
        )

    def _detect_potential_ports(self):
        all_ports = list_serial_ports()
        """
        for p in all_ports:
            print(p.description,
                  p.device,
                  None if p.vid is None else hex(p.vid),
                  None if p.pid is None else hex(p.pid),
                  )
        """
        return [
            (p.device, p.description)
            for p in all_ports
            if (p.vid, p.pid) in self.known_usb_vids_pids
            or p.description in self.known_port_descriptions
            or ("USB" in p.description and "serial" in p.description.lower())
            or "UART" in p.description
            or "DAPLink" in p.description
        ]

    @property
    def idle(self):
        return self.__idle

    @idle.setter
    def idle(self, value):
        if self.__idle != value:
            logging.debug("Serial idle %s => %s", self.__idle, value)

        self.__idle = value
        if value:
            self._has_been_idle = True

    def _fetch_builtin_modules(self):
        assert self.idle
        out, err = self._execute_and_get_response("help('modules')")
        assert err == b"", "Error was: %r" % err
        modules_str = (
            out.decode("utf-8")
            .replace("Plus any modules on the filesystem", "")
            .replace("/__init__", "")
            .replace("/", ".")
        )
        return modules_str.split()

    def _fetch_uname(self):
        assert self.idle
        res = self._execute_and_parse_value(
            "import os as __os_; print(repr(tuple(__os_.uname()))); del __os_"
        )
        return {
            "sysname": res[0],
            "nodename": res[1],
            "release": res[2],
            "version": res[3],
            "machine": res[4],
        }

    def _interrupt_to_prompt(self, clean, timeout=8):
        assert self._serial is not None

        timer = TimeHelper(timeout)

        # NB! Sometimes disconnecting and reconnecting (on macOS?)
        # too quickly causes anomalies. See CalliopeMiniProxy for more details

        for delay in [0.05, 0.5, 2.0, 3.0]:
            # Interrupt several times, because with some drivers first interrupts seem to vanish
            self._serial.reset_output_buffer()
            self._serial.write(b"\x03")  # interrupt
            self._serial.write(b"\x01")  # raw mode
            sleep(delay)
            self._discarded_bytes += self._serial.read_all()
            if self._discarded_bytes.endswith(
                FIRST_RAW_PROMPT
            ) or self._discarded_bytes.endswith(b"\r\n>"):
                break
        else:
            raise TimeoutError("Can't get to raw prompt")

        welcome_text = self._get_welcome_text_in_raw_mode(timer.time_left)

        if clean:
            self._clean_environment_during_startup(timer.time_left)

        self._finalize_repl()

        # report ready
        self._non_serial_msg_queue.put(
            ToplevelResponse(welcome_text=welcome_text.strip())
        )

        self.idle = True

    def _clean_environment_during_startup(self, time_left):
        # In MP Ctrl+D doesn't run user code, in CP it does
        self._serial.write(b"\x04")
        self._discarded_bytes = self._serial.read_until(
            [FIRST_RAW_PROMPT, RAW_PROMPT], time_left
        )

    def _get_welcome_text_in_raw_mode(self, timeout):
        timer = TimeHelper(timeout)
        # get welcome text with Ctrl+B
        self._serial.write(b"\x02")
        welcome_text = (
            self._serial.read_until(NORMAL_PROMPT, timer.time_left)
            .strip(b"\r\n >")
            .decode("utf-8", "replace")
        )
        if os.name != "nt":
            welcome_text = welcome_text.replace("\r\n", "\n")

        # Go back to raw prompt
        self._serial.write(b"\x01")
        self._serial.read_until((FIRST_RAW_PROMPT, b"\x04>"), timer.time_left)

        return welcome_text + " [backend=" + self.get_backend_name() + "]"

    def _finalize_repl(self):
        pass

    def _soft_reboot_and_run_main(self):
        if self._serial is None:
            return

        if not self.idle:
            # TODO: ignore??
            # self._serial.write(b"\r\x03")
            self.interrupt()

        get_runner()._set_state("running")
        self.idle = False
        # Need to go to normal mode. MP doesn't run user code in raw mode
        # (CP does, but it doesn't hurt to do it there as well)
        self._serial.write(b"\x02")
        self._serial.read_until(NORMAL_PROMPT)

        self._serial.write(b"\x04")

        # Returning to the raw prompt will be handled by
        # _read_next_serial_message

    def _clear_environment(self):
        assert self.idle
        # TODO: Ctrl+D in raw repl is perfect for MicroPython
        # but on CircuitPython it runs main.py

        # TODO: which is better:
        # self._execute_async(dedent("""
        #    for name in globals():
        #        if not name.startswith("__"):
        #            del globals()[name]
        # """).strip())
        # or
        out, err = self._execute_and_get_response(
            dedent(
                """
            globals().clear()
            __name__ == '__main__'
        """
            ).strip()
        )

        assert out == b""
        assert err == b""

    def _handle_serial_exception(self, e):
        logging.exception("MicroPython serial error")
        self._show_error_connect_again("\nLost connection to the device (%s)." % e)
        self.idle = False
        try:
            self._serial.close()
        except Exception:
            logging.exception("Closing serial")
        finally:
            self._serial = None

    def _execute_async(self, script):
        """Executes given MicroPython script on the device"""
        assert self._serial.buffers_are_empty()

        command_bytes = script.encode("utf-8")
        self._serial.write(command_bytes + b"\x04")
        self.idle = False

        # fetch confirmation
        ok = self._serial.read(2)
        assert ok == b"OK", "Expected OK, got %s, followed by %s" % (
            ok,
            self._serial.read_all(),
        )

    def _execute_and_get_response(self, script):
        self._execute_async(script)
        terminator = b"\x04>"
        output = self._serial.read_until(terminator)[: -len(terminator)]
        self.idle = True
        return output.split(b"\x04")

    def _execute_and_parse_value(self, script):
        out, err = self._execute_and_get_response(script)

        # TODO: report the error to stderr
        assert len(err) == 0, "Error was " + repr(err)
        return ast.literal_eval(out.strip().decode("utf-8"))

    def _execute_and_expect_empty_response(self, script):
        out, err = self._execute_and_get_response(script)

        assert len(out) == 0, "Output was " + repr(out)
        assert len(err) == 0, "Error was " + repr(err)

    def _cmd_cd(self, cmd):
        assert len(cmd.args) == 1
        path = cmd.args[0]
        if os.path.exists(path):
            self._non_serial_msg_queue.put(ToplevelResponse(cwd=path))
        else:
            self._non_serial_msg_queue.put(
                ToplevelResponse(error="Path doesn't exist: %s" % path)
            )

    def _cmd_Run(self, cmd):
        self._clear_environment()

        if not hasattr(cmd, "source"):
            assert len(cmd.args) == 1
            filename = cmd.args[0]
            if os.path.isabs(filename):
                full_filename = filename
            else:
                full_filename = os.path.join(get_workbench().get_cwd(), filename)
                cmd.script_path = full_filename

            with tokenize.open(full_filename) as fp:
                source = fp.read()
        else:
            source = cmd.source

        self._execute_async(source)

    def _cmd_execute_source(self, cmd):
        try:
            # Try to parse as expression
            ast.parse(cmd.source, mode="eval")
            # If it didn't fail then source is an expression
            msg_template = """{'message_class':'ToplevelResponse', 'value_info':(id(v), repr(v))}"""
            self._execute_async(
                "print('\\x04\\x02', [%s for v in [%s]][0])"
                % (msg_template, cmd.source.strip())
            )
        except SyntaxError:
            # source is a statement (or invalid syntax)
            self._execute_async(cmd.source)

    def _cmd_get_globals(self, cmd):
        if not get_runner().is_waiting_toplevel_command():
            return "postpone"

        try:
            if cmd.module_name == "__main__":
                self._execute_async(
                    "print('\\x04\\x02', {'message_class' : 'InlineResponse', 'command_name':'get_globals', 'module_name' : '__main__', 'globals':{x:repr(globals()[x]) for x in globals() if not x.startswith('__')}})"
                )
            else:
                self._execute_async(
                    dedent(
                        """
                    try:
                        import %(mod_name)s as __modForGlobs
                        print('\\x04\\x02', {'message_class' : 'InlineResponse', 'command_name':'get_globals', 'module_name' : '%(mod_name)s', 'globals':{name : repr(getattr(__modForGlobs, name)) for name in dir(__modForGlobs) if not name.startswith('__')}})
                        del __modForGlobs
                    except Exception as e:
                        print('\\x04\\x02', {'message_class' : 'InlineResponse', 'command_name':'get_globals', 'module_name' : '%(mod_name)s', 'globals':{}, 'error' : 'Error querying globals:\\n' + str(e)})
                    """
                        % {"mod_name": cmd.module_name}
                    )
                )
        except Exception:
            self._non_serial_msg_queue.put(
                InlineResponse(
                    command_name="get_globals",
                    module_name=cmd.module_name,
                    globals={},
                    error="Error requesting globals:\\n" + traceback.format_exc(),
                )
            )

        return None

    def _cmd_editor_autocomplete(self, cmd):
        # template for the response
        msg = InlineResponse(
            command_name="editor_autocomplete",
            source=cmd.source,
            row=cmd.row,
            column=cmd.column,
            error=None,
        )

        try:
            script = jedi.Script(
                cmd.source, cmd.row, cmd.column, sys_path=[self._get_api_stubs_path()]
            )
            completions = script.completions()
        except Exception:
            msg["error"] = "Autocomplete error"
            self._non_serial_msg_queue.put(msg)
            return

        msg["completions"] = self.filter_completions(completions)
        self._non_serial_msg_queue.put(msg)

    def filter_completions(self, completions):
        # filter out completions not applicable to MicroPython
        result = []
        for completion in completions:
            if completion.name.startswith("__"):
                continue

            parent_name = completion.parent().name
            name = completion.name
            root = completion.full_name.split(".")[0]

            # jedi proposes names from CPython builtins
            if root in self._builtins_info and name not in self._builtins_info[root]:
                continue

            if parent_name == "builtins" and name not in self._builtins_info:
                continue

            result.append({"name": name, "complete": completion.complete})

        return result

    def _cmd_shell_autocomplete(self, cmd):
        source = cmd.source

        # TODO: combine dynamic results and jedi results
        if source.strip().startswith("import ") or source.strip().startswith("from "):
            # this needs the power of jedi
            msg = InlineResponse(
                command_name="shell_autocomplete", source=cmd.source, error=None
            )

            try:
                # at the moment I'm assuming source is the code before cursor, not whole input
                lines = source.split("\n")
                script = jedi.Script(
                    source,
                    len(lines),
                    len(lines[-1]),
                    sys_path=[self._get_api_stubs_path()],
                )
                completions = script.completions()
                msg["completions"] = self.filter_completions(completions)
            except Exception:
                msg["error"] = "Autocomplete error"

            self._non_serial_msg_queue.put(msg)
        else:
            # use live data
            regex = re.search(
                r"(\w+\.)*(\w+)?$", source
            )  # https://github.com/takluyver/ubit_kernel/blob/master/ubit_kernel/kernel.py
            if regex:
                n = regex.group()
                # the response format is not the same as expected by the gui
                # but it will be transformed later
                if "." in n:
                    obj, n = n.rsplit(".", 1)
                    self._execute_async(
                        "print('\\x04\\x02', {'message_class' : 'InlineResponse', 'command_name': 'shell_autocomplete', 'match':"
                        + repr(n)
                        + ", 'source':"
                        + repr(source)
                        + ", 'names':dir("
                        + obj
                        + ")})"
                    )
                else:
                    self._execute_async(
                        "print('\\x04\\x02', {'message_class' : 'InlineResponse', 'command_name': 'shell_autocomplete', 'match':"
                        + repr(n)
                        + ", 'source':"
                        + repr(source)
                        + ", 'names':dir()})"
                    )

    def _cmd_dump_api_info(self, cmd):
        "For use during development of the plug-in"
        try:
            self._execute_and_expect_empty_response(
                dedent(
                    """
                def __print_object_atts(obj):
                    import gc
                    result = []
                    errors = []
                    for name in dir(obj):
                        try:
                            val = getattr(obj, name)
                            result.append((name, repr(val), repr(type(val))))
                        except BaseException as e:
                            errors.append("Couldn't get attr '%s' from object '%r', Err: %r" % (name, obj, e))
                    print((result, errors))
                    gc.collect()
            """
                )
            )
            for module_name in sorted(self._fetch_builtin_modules()):
                if (
                    not module_name.startswith("_")
                    # and not module_name.startswith("ada")
                    # and not module_name == "builtins"
                ):
                    # self._send_text_to_shell("Dumping " + module_name + "\n", "stdout")
                    file_name = os.path.join(
                        self._get_api_stubs_path(),
                        module_name.replace(".", "/") + ".py",
                    )
                    self._dump_module_stubs(module_name, file_name)
        finally:
            self._non_serial_msg_queue.put(ToplevelResponse())

    def _dump_module_stubs(self, module_name, file_name):
        _, err = self._execute_and_get_response("import {0}".format(module_name))
        if err:
            print("FAILED IMPORTING MODULE:", module_name, "\nErr: " + repr(err))
            return

        os.makedirs(os.path.dirname(file_name), exist_ok=True)
        with io.open(file_name, "w", encoding="utf-8", newline="\n") as fp:
            if module_name not in [
                "webrepl",
                "_webrepl",
                "gc",
                "http_client",
                "http_client_ssl",
                "http_server",
                "framebuf",
                "example_pub_button",
                "flashbdev",
            ]:
                self._dump_object_stubs(fp, module_name, "")

        _, err = self._execute_and_get_response("del {0}".format(module_name))

    def _dump_object_stubs(self, fp, object_expr, indent):
        if object_expr in [
            "docs.conf",
            "pulseio.PWMOut",
            "adafruit_hid",
            "upysh",
            # "webrepl",
            # "gc",
            # "http_client",
            # "http_server",
        ]:
            print("SKIPPING problematic name:", object_expr)
            return

        print("DUMPING", indent, object_expr)
        items, errors = self._execute_and_parse_value(
            "__print_object_atts({0})".format(object_expr)
        )

        if errors:
            print("ERRORS", errors)

        for name, rep, typ in sorted(items, key=lambda x: x[0]):
            if name.startswith("__"):
                continue

            print("DUMPING", indent, object_expr, name)
            self._send_text_to_shell("  * " + name + " : " + typ, "stdout")

            if typ in ["<class 'function'>", "<class 'bound_method'>"]:
                fp.write(indent + "def " + name + "():\n")
                fp.write(indent + "    pass\n\n")
            elif typ in ["<class 'str'>", "<class 'int'>", "<class 'float'>"]:
                fp.write(indent + name + " = " + rep + "\n")
            elif typ == "<class 'type'>" and indent == "":
                # full expansion only on toplevel
                fp.write("\n")
                fp.write(indent + "class " + name + ":\n")  # What about superclass?
                fp.write(indent + "    ''\n")
                self._dump_object_stubs(
                    fp, "{0}.{1}".format(object_expr, name), indent + "    "
                )
            else:
                # keep only the name
                fp.write(indent + name + " = None\n")

    def _cmd_cat(self, cmd):
        if len(cmd.args) != 1:
            self._send_error_to_shell("Command requires one argument")
            return

        source = cmd.args[0]
        mount = self._get_fs_mount()
        if mount is None:
            self._cat_via_serial(source)
        else:
            source = os.path.join(mount, source.strip("/"))
            self._cat_via_mount(source)

    def _cmd_lsdevice(self, cmd):
        try:
            items = self._list_files()
            out = "\n".join(items) + "\n"
            self._send_text_to_shell(out, "stdout")
        finally:
            self._non_serial_msg_queue.put(ToplevelResponse())

    def _cmd_upload(self, cmd):
        # Target is interpreted relative to the root
        if len(cmd.args) == 1:
            source = cmd.args[0]
            # target is at root
            target = os.path.basename(source)
        elif len(cmd.args) == 2:
            source = cmd.args[0]
            target = cmd.args[1]
        else:
            # TODO: test this case
            raise RuntimeError("Command requires 1 or 2 arguments")

        if not os.path.isabs(source):
            source = os.path.join(get_workbench().get_cwd(), source)

        if not os.path.isfile(source):
            raise IOError("No such file: %s" % source)

        target = target.replace("\\", "/")
        # Only prepend slash if it is known that device supports directories
        # (it's probably safe to omit slash anyway)
        if self._supports_directories() and not target.startswith("/"):
            target = "/" + target

        try:
            self._check_and_upload(source, target)
        finally:
            self._non_serial_msg_queue.put(ToplevelResponse())
            # TODO: Output confirmation ? (together with file size)
            # Or should the confirmation be given in terms of mount path?

    def _check_and_upload(self, source, target):
        # if target is a py file,
        # then give warning if source has syntax errors

        # Note that it's incomplete solution --
        # if current Python version is later than 3.5, then it may
        # accept source which gives syntax errors on MP.

        if target.endswith(".py"):
            with tokenize.open(source) as fp:
                src = fp.read()
                try:
                    ast.parse(src, source)
                except SyntaxError as e:
                    self._send_error_to_shell(
                        "%s has syntax errors:\n%s\n\nFile will not be uploaded."
                        % (source, e)
                    )
                    return
        try:
            self._upload(source, target)
        except Exception:
            self._send_error_to_shell(traceback.format_exc())

    def _upload(self, source, target):
        mount = self._get_fs_mount()
        if mount is None:
            self._upload_via_serial(source, target)
        else:
            virtual_path = os.path.join(mount, target.strip("/"))
            self._upload_via_mount(source, virtual_path)

    def _upload_via_serial(self, source, target):
        assert self.idle

        with open(source, "rb") as local:
            content = local.read()

        self._execute_and_expect_empty_response("__upf = open(%r, 'wb')" % target)

        BLOCK_SIZE = 64
        for i in range(0, len(content), BLOCK_SIZE):
            self._execute_and_expect_empty_response(
                "__upf.write(%r)" % content[i : i + BLOCK_SIZE]
            )

        self._execute_and_expect_empty_response("__upf.close()")
        self._execute_and_expect_empty_response("del __upf")

    def _upload_via_mount(self, source, target):
        with open(source, "rb") as fp:
            content = fp.read()

        try:
            with open(target, "wb") as fp:
                fp.write(content)
                # Force writes to the device to avoid data corruption
                # when user resets or plugs out the device
                os.fsync(fp)
        except OSError as e:
            self._report_upload_via_mount_error(source, target, e)
            return

    def _report_upload_via_mount_error(self, source, target, error):
        self._send_error_to_shell(
            "Couldn't write to %s\nOriginal error: %s\n\nIf the target directory exists then it may be corrupted."
            % (target, error)
        )

    def _cat_via_serial(self, source):
        try:
            out, err = self._execute_and_get_response(
                dedent(
                    """
                with open(%r, encoding='utf-8') as fp:
                    print(fp.read())
            """
                    % source
                ).strip()
            )
            if out:
                self._send_text_to_shell(
                    out.decode("utf-8", errors="replace"), "stdout"
                )
            if err:
                self._send_text_to_shell(
                    err.decode("utf-8", errors="replace"), "stderr"
                )
        except Exception:
            self._send_error_to_shell(traceback.format_exc())
        finally:
            self._non_serial_msg_queue.put(ToplevelResponse())

    def _cat_via_mount(self, source):
        try:
            with open(source, "r", encoding="UTF-8", errors="replace") as fp:
                self._send_text_to_shell(fp.read(), "stdout")

        except Exception:
            self._send_error_to_shell(traceback.format_exc())
        finally:
            self._non_serial_msg_queue.put(ToplevelResponse())

    def _list_files(self):
        mount = self._get_fs_mount()
        if mount is None:
            return self._execute_and_parse_value(
                "import os as __os_; print(__os_.listdir()); del __os_"
            )
        else:
            return os.listdir(mount)

    def _supports_directories(self):
        return None

    def _get_fs_mount_name(self):
        return None

    def _get_bootloader_mount_name(self):
        return None

    def _get_fs_mount(self):
        if self._get_fs_mount_name() is None:
            return None
        else:
            candidates = find_volumes_by_name(self._get_fs_mount_name())
            if len(candidates) == 0:
                raise RuntimeError("Could not find volume " + self._get_fs_mount_name())
            elif len(candidates) > 1:
                raise RuntimeError(
                    "Found several possible mount points: %s" % candidates
                )
            else:
                return candidates[0]

    def _read_next_serial_message(self) -> Optional[MessageFromBackend]:
        new_bytes = self._serial.read_all()

        if len(new_bytes) == 0:
            return None

        # TODO: Handle the case where part of the prompt was already published in previous message

        # Look for the first marker (EOT anywhere or NORMAL_PROMPT in the end of the seq)
        match = re.search(
            b"("
            + EOT
            + b"|"
            + NORMAL_PROMPT
            + b"$"  # Consider prompts only if they're at the end of output
            + b"|"
            + FIRST_RAW_PROMPT
            + b"$"
            + b")",
            new_bytes,
        )

        if match is None:
            # normal output (possibly partial)
            return self._read_output_message(new_bytes, False)

        elif match.start() > 0:
            # starts with block of normal output
            self._serial.unread(new_bytes[match.start() :])
            return self._read_output_message(new_bytes[: match.start()], True)

        elif match.group() == FIRST_RAW_PROMPT:
            assert new_bytes == FIRST_RAW_PROMPT
            self.idle = True
            return ToplevelResponse()

        elif match.group() == NORMAL_PROMPT:
            # Go to raw prompt
            assert new_bytes == NORMAL_PROMPT, "Got %s" % new_bytes
            return self._enter_raw_repl(True)

        else:
            assert match.group() == EOT
            assert match.start() == 0

            if len(new_bytes) == 1:
                # can't decide anything yet
                self._serial.unread(new_bytes)
                return None
            elif new_bytes[1:2] == RAW_PROMPT:
                # must be end of the response to a non-Thonny command
                # Only treat as raw prompt if it ends the output
                if new_bytes[1:] == RAW_PROMPT:
                    assert self._serial.incoming_is_empty()  # TODO: what about Ctlr-? ?
                    self.idle = True
                    return ToplevelResponse()
                else:
                    # Looks like the prompt was discarded by a soft reboot (or some other reason?)
                    # hide it and forget it
                    self._serial.unread(new_bytes[2:])
                    return None
            elif new_bytes[1:2] == THONNY_MSG_START:
                # must be followed by empty error block and raw prompt
                # read the message, following exception block and next prompt
                terminator = b"\r\n" + EOT + EOT + RAW_PROMPT
                term_loc = new_bytes.find(terminator)
                if term_loc == -1:
                    # not complete yet
                    self._serial.unread(new_bytes)
                    return None

                elif term_loc == len(new_bytes) - len(terminator):
                    # This terminator ends the bytes
                    # The normal, completed case
                    assert self._serial.incoming_is_empty()
                    msg_bytes = new_bytes[2 : -len(terminator)]
                    self.idle = True
                    return self._parse_message(msg_bytes)
                else:
                    # There is something following the terminator
                    # I guess this can be caused by interrupt
                    # This means the message is stale
                    logging.info(
                        "disregarding out of date Thonny message: %r", new_bytes
                    )
                    # Unread following stuff
                    self._serial.unread(new_bytes[term_loc + len(terminator) :])

            else:
                # exception block
                # this is followed by EOT and can/should be read in one piece
                next_eot_loc = new_bytes.find(EOT, 1)
                if next_eot_loc == -1:
                    # the block isn't complete yet
                    self._serial.unread(new_bytes)
                    return None
                else:
                    # block is complete
                    block_bytes = new_bytes[1:next_eot_loc]
                    leftover_bytes = new_bytes[next_eot_loc:]  # should be EOT + >
                    self._serial.unread(leftover_bytes)

                    if len(block_bytes) > 0:
                        # non-empty exception block
                        return BackendEvent(
                            event_type="ProgramOutput",
                            stream_name="stderr",
                            data=self.transform_output(
                                block_bytes.decode("utf-8", "replace"), "stderr"
                            ),
                        )
                    else:
                        return None

        return None

    def _parse_message(self, msg_bytes):
        msg_str = msg_bytes.decode("utf-8", "replace").strip()
        msg = ast.literal_eval(msg_str)
        assert isinstance(msg, dict)

        class_name = msg["message_class"]
        del msg["message_class"]

        assert class_name in globals()
        class_ = globals()[class_name]
        return class_(**msg)

    def _read_output_message(self, out_bytes, complete) -> Optional[BackendEvent]:

        if complete:
            out_str = out_bytes.decode("utf-8", "replace")
        else:
            # out_bytes may end with a partial utf-8 char
            while True:
                try:
                    out_str = out_bytes.decode("utf-8", "replace")
                    break
                except UnicodeDecodeError:
                    # unread last byte and try again
                    self._serial.unread(out_bytes[-1:])
                    out_bytes = out_bytes[:-1]

        if len(out_str) == 0:
            return None
        else:
            transformed = self.transform_output(out_str, "stdout")
            return BackendEvent(
                event_type="ProgramOutput", stream_name="stdout", data=transformed
            )

    def transform_output(self, s, stream_name):
        if os.name != "nt":
            #
            s = s.replace("\r\n", "\n")

        # replace "<stdin>" in error messages with script name
        if (
            stream_name == "stderr"
            and self._last_toplevel_command
            and self._last_toplevel_command.name in ["Run", "run"]
            and hasattr(self._last_toplevel_command, "script_path")
        ):
            s = s.replace('"<stdin>"', '"%s"' % self._last_toplevel_command.script_path)

        # TODO: get rid of raw prompts (may occur after soft reboot)
        # TOOD: move it to CircuitPython subclass
        return s.replace(
            "Press any key to enter the REPL. Use CTRL-D to reload.",
            "Press CTRL-C to enter the REPL. Use CTRL-D to reload.",
        )

    def _get_main_script_path(self):
        if self._supports_directories():
            return "/main.py"
        else:
            return "main.py"

    def _get_boot_script_path(self):
        if self._supports_directories():
            return "/boot.py"
        else:
            return "boot.py"

    def _get_script_path(self):
        script_path = (
            get_workbench().get_editor_notebook().get_current_editor().save_file(False)
        )
        script_path = Path(script_path)
        assert script_path.is_file(), "File not found: %s" % script_path

        filename = script_path.name

        if self._supports_directories():
            return "/%s" % filename
        else:
            return filename

    def transform_message(self, msg):
        if msg is None:
            return None

        if isinstance(msg.get("value_info", None), tuple):
            msg["value_info"] = common.ValueInfo(*msg["value_info"])

        if (
            getattr(msg, "command_name", None) == "shell_autocomplete"
            and "completions" not in msg
        ):
            names = msg["names"]
            match = msg["match"]
            del msg["names"]
            matches = [
                {"name": n, "complete": n[len(match) :]}
                for n in names
                if n.startswith(match) and not n.startswith("__")
            ]
            msg["completions"] = matches
            return msg
        else:
            return msg

    def _enter_raw_repl(self, strict):
        if strict:
            assert self._serial.buffers_are_empty()

        discarded_data = b""

        for delay in [0.01, 0.05, 0.1, 0.5]:
            self._serial.write(b"\x03")
            sleep(delay / 3)
            self._serial.write(b"\x01")
            sleep(delay)
            # Consume the raw repl introduction + prompt
            discarded_data += self._serial.read_all()
            if discarded_data.endswith(b"\r\n>"):
                self.idle = True
                return ToplevelResponse()

        self._send_error_to_shell(
            "Couldn't connect to the raw REPL. Serial output: " + str(discarded_data)
        )

        self.idle = False
        return None

    def _send_error_to_shell(self, message_text):
        self._send_text_to_shell(message_text, "stderr")

    def _send_text_to_shell(self, message_text, stream_name):
        if not message_text.endswith("\n"):
            message_text += "\n"

        self._non_serial_msg_queue.put(
            BackendEvent(
                event_type="ProgramOutput", stream_name=stream_name, data=message_text
            )
        )

    def _get_builtins_info(self):
        """
        for p in self._get_api_stubs_path():
            builtins_file = os.path.join(p, "__builtins__.py")
            if os.path.exists(builtins_file):
                return parse_api_information(builtins_file)
        """
        path = os.path.join(self._get_api_stubs_path(), "builtins.py")
        if os.path.exists(path):
            return parse_api_information(path)
        else:
            return {}

    def _get_api_stubs_path(self):
        import inspect

        return os.path.join(
            os.path.dirname(inspect.getfile(self.__class__)), "api_stubs"
        )

    @property
    def firmware_filetypes(self):
        return [("all files", ".*")]

    @property
    def micropython_upload_enabled(self):
        return self.port is not None

    def select_and_upload_micropython(self):
        firmware_path = askopenfilename(
            filetypes=self.firmware_filetypes,
            initialdir=get_workbench().get_option("run.working_directory"),
        )
        if firmware_path:
            self.upload_micropython(firmware_path)

    def upload_micropython(self, firmware_path):
        cmd = self.construct_firmware_upload_command(firmware_path)
        self.disconnect()
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        dlg = SubprocessDialog(
            get_workbench(),
            proc,
            "Uploading firmware",
            autoclose=False,
            conclusion="Done.\nNB! If opening REPL fails on first trial\nthen wait a second and try again.",
        )
        dlg.wait_window()

    def construct_firmware_upload_command(self, firmware_path):
        raise NotImplementedError()

    @property
    def known_usb_vids_pids(self):
        """Return set of pairs of USB device VID, PID"""
        return set()

    @property
    def known_port_descriptions(self):
        return set()


class MicroPythonConfigPage(BackendDetailsConfigPage):
    backend_name = None  # Will be overwritten on Workbench.add_backend

    def __init__(self, master):
        super().__init__(master)
        label = ttk.Label(
            self, text='Port (look for your device name, "USB Serial" or "UART")'
        )
        label.grid(row=0, column=0, sticky="nw")

        self._ports_by_desc = {
            p.description
            if p.device in p.description
            else p.description + " (" + p.device + ")": p.device
            for p in list_serial_ports()
        }
        self._ports_by_desc["< Try to detect automatically >"] = "auto"
        self._ports_by_desc["< None / don't connect at all >"] = None

        def port_order(p):
            _, name = p
            if name is None:
                return ""
            elif name.startswith("COM") and len(name) == 4:
                # Make one-digit COM ports go before COM10
                return name.replace("COM", "COM0")
            else:
                return name

        # order by port, auto first
        port_descriptions = [
            key for key, _ in sorted(self._ports_by_desc.items(), key=port_order)
        ]

        self._port_desc_variable = create_string_var(self.current_port_desc)
        self._port_combo = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=self._port_desc_variable,
            values=port_descriptions,
        )
        self._port_combo.state(["!disabled", "readonly"])

        self._port_combo.grid(row=1, column=0, sticky="new")
        self.columnconfigure(0, weight=1)

    @property
    def current_port_desc(self):
        name = get_workbench().get_option(self.backend_name + ".port")
        for desc in self._ports_by_desc:
            if self._ports_by_desc[desc] == name:
                return desc

        return ""

    def should_restart(self):
        return self._port_desc_variable.modified  # pylint: disable=no-member

    def apply(self):
        if not self._port_desc_variable.modified:  # pylint: disable=no-member
            return

        else:
            port_desc = self._port_desc_variable.get()
            port_name = self._ports_by_desc[port_desc]
            get_workbench().set_option(self.backend_name + ".port", port_name)


class GenericMicroPythonProxy(MicroPythonProxy):
    @property
    def known_usb_vids_pids(self):
        """Return set of pairs of USB device (VID, PID)"""
        return {
            # Generic MicroPython Board, see http://pid.codes/org/MicroPython/
            (0x1209, 0xADDA)
        }


class GenericMicroPythonConfigPage(MicroPythonConfigPage):
    pass


class SerialHelper:
    """Utility class for using Serial connection
    
    Uses background thread to read from serial as soon as possible
    to avoid loss of data (because buffer overflow or the device discarding 
    unread data).

    Allows writing with delays after each n bytes.
    
    Allows unreading data.    
    """

    def __init__(self, port, baudrate):
        self._read_queue = Queue()  # populated by reader thread
        self._read_buffer = bytearray()  # used for unreading and postponing bytes
        self._serial = serial.Serial(port, baudrate=baudrate, timeout=None)
        self.num_bytes_received = 0
        self._error = None

        self._reading_thread = threading.Thread(target=self._listen_serial, daemon=True)
        self._reading_thread.start()

    def read(self, size, timeout=1):
        if timeout == 0:
            raise TimeoutError()

        timer = TimeHelper(timeout)

        while len(self._read_buffer) < size:
            self._check_for_error()

            try:
                self._read_buffer.extend(self._read_queue.get(True, timer.time_left))
            except queue.Empty:
                raise TimeoutError(
                    "Reaction timeout. Bytes read: %s" % self._read_buffer
                )

        try:
            data = self._read_buffer[:size]
            return data
        finally:
            del self._read_buffer[:size]

    def read_until(self, terminators, timeout=2):
        if timeout == 0:
            raise TimeoutError()

        timer = TimeHelper(timeout)
        if not isinstance(terminators, (set, list, tuple)):
            terminators = [terminators]

        terminator = None
        while True:
            self._check_for_error()

            found = False
            for terminator in terminators:
                if terminator in self._read_buffer:
                    found = True
                    break
            if found:
                break

            try:
                data = self._read_queue.get(True, timer.time_left)
                assert len(data) > 0
                self._read_buffer.extend(data)
            except queue.Empty:
                raise TimeoutError(
                    "Reaction timeout. Bytes read: %s" % self._read_buffer
                )

        assert terminator is not None
        size = self._read_buffer.index(terminator) + len(terminator)

        try:
            data = self._read_buffer[:size]
            return data
        finally:
            del self._read_buffer[:size]

    def read_all(self):
        while not self._read_queue.empty():
            self._read_buffer.extend(self._read_queue.get(True))

        if len(self._read_buffer) == 0:
            self._check_for_error()

        try:
            return self._read_buffer
        finally:
            self._read_buffer = bytearray()

    def _check_for_error(self):
        if self._error:
            raise SerialException("EOF")

    def unread(self, data):
        self._read_buffer = data + self._read_buffer

    def write(self, data, block_size=32, delay=0.01):
        for i in range(0, len(data), block_size):
            block = data[i : i + block_size]
            # self._log_data(b"[" + block + b"]")
            size = self._serial.write(block)
            assert size == len(block)
            time.sleep(delay)
        return len(data)

    def _listen_serial(self):
        "NB! works in background thread"
        try:
            while True:
                b = self._serial.read(1)  # To avoid busy loop
                if len(b) == 0:
                    self._error = "EOF"
                    # print("LISTEN EOFFFFFFFFFF")
                    break
                data = b + self._serial.read_all()
                self.num_bytes_received += len(data)
                self._read_queue.put(data)
                # self._log_data(data)

        except SerialException as e:
            logging.exception("Error while reading from serial")
            self._error = str("Serial reading error: %s" % e)

    def _log_data(self, data):
        print(
            data.decode("Latin-1")
            .replace("\r\n", "\n")
            .replace("\x01", "①")
            .replace("\x02", "②")
            .replace("\x03", "③")
            .replace("\x04", "④"),
            end="",
        )

    def incoming_is_empty(self):
        return (
            self._serial.in_waiting == 0
            and self._read_queue.empty()
            and len(self._read_buffer) == 0
        )

    def outgoing_is_empty(self):
        return self._serial.out_waiting == 0

    def buffers_are_empty(self):
        return self.incoming_is_empty() and self.outgoing_is_empty()

    def reset_input_buffer(self):
        return self.read_all()

    def reset_output_buffer(self):
        self._serial.reset_output_buffer()

    def close(self):
        if self._serial is not None:
            try:
                self._serial.cancel_read()
                self._reading_thread.join()
            finally:
                try:
                    self._serial.close()
                    self._serial = None
                except Exception:
                    logging.exception("Couldn't close serial")


class TimeHelper:
    def __init__(self, time_allowed):
        self.start_time = time.time()
        self.time_allowed = time_allowed

    @property
    def time_spent(self):
        return time.time() - self.start_time

    @property
    def time_left(self):
        return max(self.time_allowed - self.time_spent, 0)


def parse_api_information(file_path):
    with tokenize.open(file_path) as fp:
        source = fp.read()

    tree = ast.parse(source)

    defs = {}

    # TODO: read also docstrings ?

    for toplevel_item in tree.body:
        if isinstance(toplevel_item, ast.ClassDef):
            class_name = toplevel_item.name
            member_names = []
            for item in toplevel_item.body:
                if isinstance(item, ast.FunctionDef):
                    member_names.append(item.name)
                elif isinstance(item, ast.Assign):
                    # TODO: check Python 3.4
                    "TODO: item.targets[0].id"

            defs[class_name] = member_names

    return defs


def list_serial_ports():
    # serial.tools.list_ports.comports() can be too slow
    # because os.path.islink can be too slow (https://github.com/pyserial/pyserial/pull/303)
    # Workarond: temporally patch os.path.islink
    try:
        old_islink = os.path.islink
        if platform.system() == "Windows":
            os.path.islink = lambda _: False
        return list(serial.tools.list_ports.comports())
    finally:
        os.path.islink = old_islink


def add_micropython_backend(name, proxy_class, description, config_page):
    get_workbench().set_default(name + ".port", "auto")
    get_workbench().add_backend(name, proxy_class, description, config_page)


def load_plugin():
    add_micropython_backend(
        "GenericMicroPython",
        GenericMicroPythonProxy,
        "MicroPython on a generic device",
        GenericMicroPythonConfigPage,
    )

    def _upload_as(target_provider_method):
        source_path = (
            get_workbench().get_editor_notebook().get_current_editor().save_file(False)
        )
        if source_path is None:
            return

        proxy = get_runner().get_backend_proxy()
        assert isinstance(proxy, MicroPythonProxy)

        if os.path.isabs(source_path):
            source_path = os.path.relpath(source_path, get_workbench().get_cwd())

        target = getattr(proxy, target_provider_method)()
        get_shell().submit_magic_command(["%upload", source_path, target])

    def _cat(source_provider_method):
        proxy = get_runner().get_backend_proxy()
        assert isinstance(proxy, MicroPythonProxy)

        source = getattr(proxy, source_provider_method)()
        get_shell().submit_magic_command(["%cat", source])

    def _upload_as_main_script():
        _upload_as("_get_main_script_path")

    def _upload_as_boot_script():
        _upload_as("_get_boot_script_path")

    def _upload_script():
        _upload_as("_get_script_path")

    def _cat_main_script():
        _cat("_get_main_script_path")

    def _cat_boot_script():
        _cat("_get_boot_script_path")

    def soft_reboot():
        proxy = get_runner().get_backend_proxy()
        if hasattr(proxy, "_soft_reboot_and_run_main"):
            return proxy._soft_reboot_and_run_main()
        return None

    def soft_reboot_enabled():
        proxy = get_runner().get_backend_proxy()
        return (
            proxy
            and proxy.is_functional()
            and hasattr(proxy, "_soft_reboot_and_run_main")
        )

    def disconnect():
        proxy = get_runner().get_backend_proxy()
        assert hasattr(proxy, "disconnect")
        proxy.disconnect()

    def disconnect_enabled():
        proxy = get_runner().get_backend_proxy()
        return hasattr(proxy, "disconnect")

    def file_commands_enabled():
        proxy = get_runner().get_backend_proxy()
        return (
            isinstance(proxy, MicroPythonProxy)
            and get_workbench().get_editor_notebook().get_current_editor() is not None
            and get_runner().is_waiting_toplevel_command()
        )

    get_workbench().add_command(
        "softreboot",
        "device",
        "Soft reboot",
        soft_reboot,
        soft_reboot_enabled,
        group=10,
        default_sequence="<Control-d>",
        extra_sequences=["<<CtrlDInText>>"],
    )

    get_workbench().add_command(
        "uploadmainscript",
        "device",
        "Upload current script as main script",
        _upload_as_main_script,
        tester=file_commands_enabled,
        default_sequence="<Control-u>",
        group=20,
    )

    get_workbench().add_command(
        "uploadbootscript",
        "device",
        "Upload current script as boot script",
        _upload_as_boot_script,
        tester=file_commands_enabled,
        group=20,
    )

    get_workbench().add_command(
        "uploadscript",
        "device",
        "Upload current script with current name",
        _upload_script,
        tester=file_commands_enabled,
        group=20,
    )

    get_workbench().add_command(
        "catmainscript",
        "device",
        "Show device's main script",
        _cat_main_script,
        tester=file_commands_enabled,
        group=20,
    )

    get_workbench().add_command(
        "catbootscript",
        "device",
        "Show device's boot script",
        _cat_boot_script,
        tester=file_commands_enabled,
        group=20,
    )

    get_workbench().add_command(
        "disconnectserial",
        "device",
        "Close serial connection",
        disconnect,
        disconnect_enabled,
        group=100,
    )



import astroid

from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker


class FileNamingChecker(BaseChecker):
    __implements__ = IAstroidChecker

    name = "file-naming"
    priority = -1
    msgs = {
        "W7401": (
            "Script shadows a library module.",
            "non-unique-returns",
            "All constants returned in a function should be unique.",
        )
    }
    options = (
        (
            "ignore-ints",
            {
                "default": False,
                "type": "yn",
                "metavar": "<y_or_n>",
                "help": "Allow returning non-unique integers",
            },
        ),
    )



from thonny import get_workbench
from thonny.workbench import SyntaxThemeSettings


def default_light() -> SyntaxThemeSettings:
    default_fg = "black"
    default_bg = "#fdfdfd"
    light_fg = "DarkGray"
    string_fg = "DarkGreen"
    open_string_bg = "#c3f9d3"

    return {
        "TEXT": {
            "foreground": default_fg,
            "insertbackground": default_fg,
            "background": default_bg,
        },
        "GUTTER": {"foreground": "#999999", "background": "#e0e0e0"},
        "breakpoint": {"foreground": "crimson"},
        "current_line": {"background": "#f5f5f5"},
        "definition": {"foreground": "DarkBlue", "font": "BoldEditorFont"},
        "string": {"foreground": string_fg},
        "string3": {"foreground": string_fg},
        "open_string": {"foreground": string_fg, "background": open_string_bg},
        "open_string3": {"foreground": string_fg, "background": open_string_bg},
        "keyword": {"foreground": "#7f0055", "font": "BoldEditorFont"},
        "builtin": {"foreground": "#7f0055"},
        "number": {"foreground": "#B04600"},
        "comment": {"foreground": light_fg},
        "prompt": {"foreground": "purple", "font": "BoldEditorFont"},
        "magic": {"foreground": light_fg},
        "stdin": {"foreground": "Blue"},
        "stdout": {"foreground": "Black"},
        "stderr": {"foreground": "Red"},
        "value": {"foreground": "DarkBlue"},
        "hyperlink": {"foreground": "#3A66DD", "underline": True},
        # paren matcher
        "surrounding_parens": {"foreground": "Blue", "font": "BoldEditorFont"},
        "unclosed_expression": {"background": "LightGray"},
        # find/replace
        "found": {"foreground": "blue", "underline": True},
        "current_found": {"foreground": "white", "background": "red"},
        "matched_name": {"background": "#e6ecfe"},
        "local_name": {"font": "ItalicEditorFont"},
        # debugger
        "active_focus": {"background": "#F8FC9A", "borderwidth": 1, "relief": "solid"},
        "suspended_focus": {"background": "", "borderwidth": 1, "relief": "solid"},
        "completed_focus": {
            "background": "#BBEDB2",
            "borderwidth": 1,
            "relief": "flat",
        },
        "exception_focus": {
            "background": "#FFBFD6",
            "borderwidth": 1,
            "relief": "solid",
        },
        "expression_box": {"background": "#DCEDF2", "foreground": default_fg},
    }


def default_dark() -> SyntaxThemeSettings:
    default_fg = "#B3B3B3"
    string_fg = "#8DC76F"
    open_string_bg = "#224533"

    # s.configure("Local.Code", foreground="#BCCAE8")
    # s.configure("MatchedName.Code", background="#193022")

    return {
        "TEXT": {
            "foreground": default_fg,
            "insertbackground": default_fg,
            "background": "#2d2d2d",
        },
        "GUTTER": {"foreground": "#606060", "background": "#323232"},
        "breakpoint": {"foreground": "pink"},
        "current_line": {"background": "#363636"},
        "sel": {"foreground": "#eeeeee", "background": "#6E6E6E"},
        "definition": {"foreground": default_fg},
        "string": {"foreground": string_fg},
        "string3": {"foreground": string_fg},
        "open_string": {"foreground": string_fg, "background": open_string_bg},
        "open_string3": {"foreground": string_fg, "background": open_string_bg},
        "builtin": {"foreground": "#A9B1C9"},
        "keyword": {"foreground": "#A9B1C9", "font": "BoldEditorFont"},
        "number": {"foreground": "#FFCABF"},
        "comment": {"foreground": "#D4D44E"},
        # shell
        "prompt": {"foreground": "#5BEBBB", "font": "BoldEditorFont"},
        "magic": {"foreground": "pink"},
        "stdin": {"foreground": "LightBlue"},
        "stdout": {"foreground": "LightGray"},
        "stderr": {"foreground": "#EB5B83"},
        "value": {"foreground": "#EBEB5B"},
        "hyperlink": {"foreground": "#DC5BEB", "underline": True},
        # paren matcher
        "surrounding_parens": {"foreground": "#F0995B", "font": "BoldEditorFont"},
        "unclosed_expression": {"background": "#000000"},
        # find/replace
        "found": {"underline": True},
        "current_found": {"foreground": "white", "background": "red"},
        "matched_name": {"background": "#474747"},
        "local_name": {"font": "ItalicEditorFont"},
        # debugger
        "active_focus": {"background": "#807238", "borderwidth": 1, "relief": "solid"},
        "suspended_focus": {"background": "", "borderwidth": 1, "relief": "solid"},
        "completed_focus": {
            "background": "#807238",
            "borderwidth": 1,
            "relief": "flat",
        },
        "exception_focus": {
            "background": "#FFBFD6",
            "borderwidth": 1,
            "relief": "solid",
        },
        "expression_box": {"background": "#506E67", "foreground": default_fg},
    }


def default_dark_green() -> SyntaxThemeSettings:
    open_string_bg = "#453B22"

    return {
        "TEXT": {"background": "#273627"},
        "GUTTER": {"background": "#33402F"},
        "current_line": {"background": "#2E402E"},
        "sel": {"background": "#6E6E6E"},
        "unclosed_expression": {"background": "#0F1F15"},
        "open_string": {"background": open_string_bg},
        "open_string3": {"background": open_string_bg},
        "keyword": {"foreground": "#88CFB6", "font": "BoldEditorFont"},
        "builtin": {"foreground": "#88CFB6"},
        # debugger
        "active_focus": {"background": "#807238"},
        "completed_focus": {"background": "#807238"},
        "exception_focus": {"background": "#FFBFD6"},
        "expression_box": {"background": "#506E67"},
    }


def default_dark_blue() -> SyntaxThemeSettings:
    open_string_bg = "#224533"
    return {
        "TEXT": {"background": "#272936"},
        "GUTTER": {"background": "#2F3640"},
        "current_line": {"background": "#2D3040"},
        "sel": {"background": "#6E6E6E"},
        "unclosed_expression": {"background": "#100B21"},
        "open_string": {"background": open_string_bg},
        "open_string3": {"background": open_string_bg},
        "keyword": {"foreground": "#8899CF", "font": "BoldEditorFont"},
        "builtin": {"foreground": "#8899CF"},
        # debugger
        "active_focus": {"background": "#807238"},
        "completed_focus": {"background": "#807238"},
        "exception_focus": {"background": "#FFBFD6"},
        "expression_box": {"background": "#506E67"},
    }


def idle_classic() -> SyntaxThemeSettings:
    string_fg = "#00aa00"
    return {
        "TEXT": {
            "foreground": "black",
            "insertbackground": "black",
            "background": "white",
        },
        "GUTTER": {"foreground": "gray", "background": "#efefef"},
        "sel": {"foreground": "black", "background": "gray"},
        "number": {"foreground": "black"},
        "definition": {"foreground": "#0000ff", "font": "EditorFont"},
        "string": {"foreground": string_fg},
        "string3": {"foreground": string_fg},
        "open_string": {"foreground": string_fg},
        "open_string3": {"foreground": string_fg},
        "keyword": {"foreground": "#ff7700", "font": "EditorFont"},
        "builtin": {"foreground": "#900090"},
        "comment": {"foreground": "#dd0000"},
        "prompt": {"foreground": "#770000"},
        "stdin": {"foreground": "black"},
        "stdout": {"foreground": "Blue"},
        "value": {"foreground": "Blue"},
        "stderr": {"foreground": "Red"},
        "found": {"foreground": "", "underline": True},
        "current_found": {"foreground": "white", "background": "black"},
    }


def idle_dark() -> SyntaxThemeSettings:
    normal_fg = "white"
    string_fg = "#02ff02"

    return {
        "TEXT": {
            "foreground": normal_fg,
            "insertbackground": normal_fg,
            "background": "#002240",
        },
        "sel": {"foreground": "#FFFFFF", "background": "#7e7e7e"},
        "number": {"foreground": normal_fg},
        "definition": {"foreground": "#5e5eff", "font": "EditorFont"},
        "string": {"foreground": string_fg},
        "string3": {"foreground": string_fg},
        "open_string": {"foreground": string_fg},
        "open_string3": {"foreground": string_fg},
        "keyword": {"foreground": "#ff8000", "font": "EditorFont"},
        "builtin": {"foreground": "#ff00ff"},
        "comment": {"foreground": "#dd0000"},
        "prompt": {"foreground": "#ff4d4d"},
        "stdin": {"foreground": normal_fg},
        "stdout": {"foreground": "#c2d1fa"},
        "value": {"foreground": "#c2d1fa"},
        "stderr": {"foreground": "#ffb3b3"},
        "found": {"foreground": "", "underline": True},
        "current_found": {"foreground": "#002240", "background": "#fbfbfb"},
    }


def desert_sunset() -> SyntaxThemeSettings:
    normal_fg = "#f0e68c"
    string_fg = "#ffa0a0"

    return {
        "TEXT": {
            "foreground": normal_fg,
            "insertbackground": normal_fg,
            "background": "#333333",
        },
        "GUTTER": {"foreground": "gray", "background": "#404040"},
        "sel": {"foreground": "#000000", "background": "gray"},
        "number": {"foreground": normal_fg},
        "definition": {"foreground": "#98fb98"},
        "string": {"foreground": string_fg},
        "string3": {"foreground": string_fg},
        "open_string": {"foreground": string_fg},
        "open_string3": {"foreground": string_fg},
        "keyword": {"foreground": "#cc6600"},
        "builtin": {"foreground": "#519e51"},
        "comment": {"foreground": "#87ceeb"},
        "prompt": {"foreground": "#87ceeb"},
        "stdin": {"foreground": normal_fg},
        "stdout": {"foreground": "#eeeeee"},
        "value": {"foreground": "#eeeeee"},
        "stderr": {"foreground": "#ff3e40"},
        "found": {"foreground": "", "underline": True},
        "current_found": {"foreground": "#ffffff", "background": "#333333"},
    }


def zenburn() -> SyntaxThemeSettings:
    # https://github.com/mig/gedit-themes/blob/master/zenburn.xml
    # https://github.com/trusktr/gedit-color-schemes/blob/master/gtksourceview-3.0/styles/zenburn.xml
    normal_fg = "#dcdccc"
    string_fg = "#cc9393"

    return {
        "TEXT": {
            "foreground": normal_fg,
            "insertbackground": normal_fg,
            "background": "#3f3f3f",
        },
        "GUTTER": {"foreground": "#7f8f8f", "background": "#464646"},
        "current_line": {"background": "#4A4A4A"},
        "sel": {"foreground": "white", "background": "#506070"},
        "number": {"foreground": "#8cd0d3"},
        "definition": {"foreground": "#f4a020", "font": "BoldEditorFont"},
        "string": {"foreground": string_fg},
        "string3": {"foreground": string_fg},
        "open_string": {"foreground": string_fg},
        "open_string3": {"foreground": string_fg},
        "keyword": {"foreground": "#f0dfaf", "font": "BoldEditorFont"},
        "builtin": {"foreground": "#efef8f"},
        "comment": {"foreground": "#7f9f7f"},
        "prompt": {"foreground": "#87ceeb"},
        "stdin": {"foreground": normal_fg},
        "stdout": {"foreground": "#eeeeee"},
        "value": {"foreground": "#eeeeee"},
        "stderr": {"foreground": "#ff3e40"},
        # paren matcher
        "surrounding_parens": {"foreground": "white", "font": "BoldEditorFont"},
    }


def load_plugin() -> None:
    get_workbench().add_syntax_theme("Default Light", None, default_light)
    get_workbench().add_syntax_theme("Default Dark", None, default_dark)
    get_workbench().add_syntax_theme(
        "Default Dark Green", "Default Dark", default_dark_green
    )
    get_workbench().add_syntax_theme(
        "Default Dark Blue", "Default Dark", default_dark_blue
    )
    get_workbench().add_syntax_theme("Desert Sunset", "Default Dark", desert_sunset)
    get_workbench().add_syntax_theme("Zenburn", "Default Dark", zenburn)
    get_workbench().add_syntax_theme("IDLE Classic", "Default Light", idle_classic)

    # Comments in IDLE Dark really hurt the eyes
    # get_workbench().add_syntax_theme("IDLE Dark", "Default Dark", idle_dark)

    get_workbench().set_default("view.syntax_theme", "Default Light")



import logging
import tkinter as tk
from tkinter import ttk

from thonny import get_workbench
from thonny.config_ui import ConfigurationPage


class EditorConfigurationPage(ConfigurationPage):
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)

        try:
            self.add_checkbox("view.name_highlighting", "Highlight matching names")
        except Exception:
            # name matcher may have been disabled
            logging.warning("Couldn't create name matcher checkbox")

        try:
            self.add_checkbox("view.locals_highlighting", "Highlight local variables")
        except Exception:
            # locals highlighter may have been disabled
            logging.warning("Couldn't create name locals highlighter checkbox")

        self.add_checkbox("view.paren_highlighting", "Highlight parentheses")
        self.add_checkbox("view.syntax_coloring", "Highlight syntax elements")
        self.add_checkbox(
            "view.highlight_current_line",
            "Highlight current line (requires reopening the editor)",
        )

        self.add_checkbox(
            "edit.tab_complete_in_editor",
            "Allow code completion with Tab-key in editors",
            columnspan=2,
            pady=(20, 0),
        )
        self.add_checkbox(
            "edit.tab_complete_in_shell",
            "Allow code completion with Tab-key in Shell",
            columnspan=2,
        )

        self.add_checkbox("view.show_line_numbers", "Show line numbers", pady=(20, 0))
        self._line_length_var = get_workbench().get_variable(
            "view.recommended_line_length"
        )
        label = ttk.Label(
            self,
            text="Recommended maximum line length\n(Set to 0 to turn off margin line)",
        )
        label.grid(row=20, column=0, sticky=tk.W)
        self._line_length_combo = ttk.Combobox(
            self,
            width=4,
            exportselection=False,
            textvariable=self._line_length_var,
            state="readonly",
            values=[0, 60, 70, 80, 90, 100, 110, 120],
        )
        self._line_length_combo.grid(row=20, column=1, sticky=tk.E)

        self.columnconfigure(0, weight=1)

    def apply(self):
        ConfigurationPage.apply(self)
        get_workbench().get_editor_notebook().update_appearance()


def load_plugin() -> None:
    get_workbench().add_configuration_page("Editor", EditorConfigurationPage)



import json
import os.path
import time
import tkinter as tk
import zipfile
from datetime import datetime

from thonny import THONNY_USER_DIR, get_workbench
from thonny.shell import ShellView
from thonny.ui_utils import asksaveasfilename
from thonny.workbench import WorkbenchEvent


class EventLogger:
    def __init__(self, filename):
        self._filename = filename
        self._init_logging()
        self._init_commands()

    def _init_commands(self):
        get_workbench().add_command(
            "export_usage_logs",
            "tools",
            "Export usage logs...",
            self._cmd_export,
            group=110,
        )

    def _init_logging(self):
        self._events = []

        wb = get_workbench()
        wb.bind("WorkbenchClose", self._on_worbench_close, True)

        for sequence in [
            "<<Undo>>",
            "<<Redo>>",
            "<<Cut>>",
            "<<Copy>>",
            "<<Paste>>",
            # "<<Selection>>",
            # "<Key>",
            # "<KeyRelease>",
            "<Button-1>",
            "<Button-2>",
            "<Button-3>",
        ]:
            self._bind_all(sequence)

        for sequence in [
            "UiCommandDispatched",
            "MagicCommand",
            "Open",
            "Save",
            "SaveAs",
            "NewFile",
            "EditorTextCreated",
            # "ShellTextCreated", # Too bad, this event happens before event_logging is loaded
            "ShellCommand",
            "ShellInput",
            "ShowView",
            "HideView",
            "TextInsert",
            "TextDelete",
        ]:
            self._bind_workbench(sequence)

        self._bind_workbench("<FocusIn>", True)
        self._bind_workbench("<FocusOut>", True)

        ### log_user_event(KeyPressEvent(self, e.char, e.keysym, self.text.index(tk.INSERT)))

        # TODO: if event data includes an Editor, then look up also text id

    def _bind_workbench(self, sequence, only_workbench_widget=False):
        def handle(event):
            if not only_workbench_widget or event.widget == get_workbench():
                self._log_event(sequence, event)

        get_workbench().bind(sequence, handle, True)

    def _bind_all(self, sequence):
        def handle(event):
            self._log_event(sequence, event)

        tk._default_root.bind_all(sequence, handle, True)

    def _extract_interesting_data(self, event, sequence):
        attributes = vars(event)

        # generate some new attributes
        if "text_widget" not in attributes:
            if "editor" in attributes:
                attributes["text_widget"] = attributes["editor"].get_text_widget()

            if "widget" in attributes and isinstance(attributes["widget"], tk.Text):
                attributes["text_widget"] = attributes["widget"]

        if "text_widget" in attributes:
            widget = attributes["text_widget"]
            if isinstance(widget.master, ShellView):
                attributes["text_widget_context"] = "shell"

        # select attributes
        data = {}
        for name in attributes:
            # skip some attributes
            if (
                name.startswith("_")
                or isinstance(event, WorkbenchEvent)
                and name in ["update", "setdefault"]
                or isinstance(event, tk.Event)
                and name not in ["widget", "text_widget", "text_widget_context"]
            ):
                continue

            value = attributes[name]

            if isinstance(value, (tk.BaseWidget, tk.Tk)):
                data[name + "_id"] = id(value)
                data[name + "_class"] = value.__class__.__name__

            elif isinstance(value, (str, int, float)):
                data[name] = value

            else:
                data[name] = repr(value)

        return data

    def _get_log_dir(self):
        return os.path.dirname(self._filename)

    def _cmd_export(self):

        filename = asksaveasfilename(
            filetypes=[("Zip-files", ".zip"), ("all files", ".*")],
            defaultextension=".zip",
            initialdir=get_workbench().get_cwd(),
            initialfile=time.strftime("ThonnyUsageLogs_%Y-%m-%d.zip"),
        )

        if not filename:
            return

        log_dir = self._get_log_dir()

        with zipfile.ZipFile(filename, "w", compression=zipfile.ZIP_DEFLATED) as zipf:
            for item in os.listdir(log_dir):
                if item.endswith(".txt") or item.endswith(".zip"):
                    zipf.write(os.path.join(log_dir, item), arcname=item)

    def _log_event(self, sequence, event):
        data = self._extract_interesting_data(event, sequence)
        data["sequence"] = sequence
        data["time"] = datetime.now().isoformat()
        self._events.append(data)

    def _on_worbench_close(self, event=None):
        with open(self._filename, encoding="UTF-8", mode="w") as fp:
            json.dump(self._events, fp, indent="    ")

        self._check_compress_logs()

    def _check_compress_logs(self):
        # if uncompressed logs have grown over 10MB,
        # compress these into new zipfile

        log_dir = self._get_log_dir()
        total_size = 0
        uncompressed_files = []
        for item in os.listdir(log_dir):
            if item.endswith(".txt"):
                full_name = os.path.join(log_dir, item)
                total_size += os.stat(full_name).st_size
                uncompressed_files.append((item, full_name))

        if total_size > 10 * 1024 * 1024:
            zip_filename = _generate_timestamp_file_name("zip")
            with zipfile.ZipFile(
                zip_filename, "w", compression=zipfile.ZIP_DEFLATED
            ) as zipf:
                for item, full_name in uncompressed_files:
                    zipf.write(full_name, arcname=item)

            for _, full_name in uncompressed_files:
                os.remove(full_name)


def _generate_timestamp_file_name(extension):
    # generate log filename
    folder = os.path.expanduser(os.path.join(THONNY_USER_DIR, "user_logs"))
    if not os.path.exists(folder):
        os.makedirs(folder)

    for i in range(100):
        filename = os.path.join(
            folder, time.strftime("%Y-%m-%d_%H-%M-%S_{}.{}".format(i, extension))
        )
        if not os.path.exists(filename):
            return filename

    raise RuntimeError()


def load_plugin() -> None:
    filename = _generate_timestamp_file_name("txt")
    # create logger
    EventLogger(filename)



# -*- coding: utf-8 -*-
import collections
import logging
import os
import platform
import re
import shutil
import signal
import subprocess
import textwrap
import threading
import time
import tkinter as tk
import traceback
from tkinter import filedialog, messagebox, ttk
from typing import Callable, List, Optional, Tuple, Union  # @UnusedImport

from thonny import get_workbench, misc_utils, tktextext
from thonny.common import TextRange
from thonny.misc_utils import running_on_linux, running_on_mac_os, running_on_windows
from thonny.tktextext import TweakableText
import sys


class CustomMenubar(ttk.Frame):
    def __init__(self, master):
        ttk.Frame.__init__(self, master, style="CustomMenubar.TFrame")
        self._menus = []
        self._opened_menu = None

        ttk.Style().map(
            "CustomMenubarLabel.TLabel",
            background=[
                ("!active", lookup_style_option("Menubar", "background", "gray")),
                (
                    "active",
                    lookup_style_option("Menubar", "activebackground", "LightYellow"),
                ),
            ],
            foreground=[
                ("!active", lookup_style_option("Menubar", "foreground", "black")),
                ("active", lookup_style_option("Menubar", "activeforeground", "black")),
            ],
        )

    def add_cascade(self, label, menu):
        label_widget = ttk.Label(
            self,
            style="CustomMenubarLabel.TLabel",
            text=label,
            padding=[6, 3, 6, 2],
            font="TkDefaultFont",
        )

        if len(self._menus) == 0:
            padx = (6, 0)
        else:
            padx = 0

        label_widget.grid(row=0, column=len(self._menus), padx=padx)

        def enter(event):
            label_widget.state(("active",))

            # Don't know how to open this menu when another menu is open
            # another tk_popup just doesn't work unless old menu is closed by click or Esc
            # https://stackoverflow.com/questions/38081470/is-there-a-way-to-know-if-tkinter-optionmenu-dropdown-is-active
            # unpost doesn't work in Win and Mac: https://www.tcl.tk/man/tcl8.5/TkCmd/menu.htm#M62
            # print("ENTER", menu, self._opened_menu)
            if self._opened_menu is not None:
                self._opened_menu.unpost()
                click(event)

        def leave(event):
            label_widget.state(("!active",))

        def click(event):
            try:
                # print("Before")
                self._opened_menu = menu
                menu.tk_popup(
                    label_widget.winfo_rootx(),
                    label_widget.winfo_rooty() + label_widget.winfo_height(),
                )
            finally:
                # print("After")
                self._opened_menu = None

        label_widget.bind("<Enter>", enter, True)
        label_widget.bind("<Leave>", leave, True)
        label_widget.bind("<1>", click, True)
        self._menus.append(menu)


class AutomaticPanedWindow(tk.PanedWindow):
    """
    Enables inserting panes according to their position_key-s.
    Automatically adds/removes itself to/from its master AutomaticPanedWindow.
    Fixes some style glitches.
    """

    def __init__(self, master, position_key=None, preferred_size_in_pw=None, **kwargs):

        tk.PanedWindow.__init__(self, master, **kwargs)
        self._pane_minsize = 100
        self.position_key = position_key
        self._restoring_pane_sizes = False

        self._last_window_size = (0, 0)
        self._full_size_not_final = True
        self._configure_binding = self.bind("<Configure>", self._on_window_resize, True)
        self._update_appearance_binding = self.bind(
            "<<ThemeChanged>>", self._update_appearance, True
        )
        self.bind("<B1-Motion>", self._on_mouse_dragged, True)
        self._update_appearance()

        # should be in the end, so that it can be detected when
        # constructor hasn't completed yet
        self.preferred_size_in_pw = preferred_size_in_pw

    def insert(self, pos, child, **kw):
        kw.setdefault("minsize", self._pane_minsize)

        if pos == "auto":
            # According to documentation I should use self.panes()
            # but this doesn't return expected widgets
            for sibling in sorted(
                self.pane_widgets(),
                key=lambda p: p.position_key if hasattr(p, "position_key") else 0,
            ):
                if (
                    not hasattr(sibling, "position_key")
                    or sibling.position_key == None
                    or sibling.position_key > child.position_key
                ):
                    pos = sibling
                    break
            else:
                pos = "end"

        if isinstance(pos, tk.Widget):
            kw["before"] = pos

        self.add(child, **kw)

    def add(self, child, **kw):
        kw.setdefault("minsize", self._pane_minsize)

        tk.PanedWindow.add(self, child, **kw)
        self._update_visibility()
        self._check_restore_preferred_sizes()

    def remove(self, child):
        tk.PanedWindow.remove(self, child)
        self._update_visibility()
        self._check_restore_preferred_sizes()

    def forget(self, child):
        tk.PanedWindow.forget(self, child)
        self._update_visibility()
        self._check_restore_preferred_sizes()

    def destroy(self):
        self.unbind("<Configure>", self._configure_binding)
        self.unbind("<<ThemeChanged>>", self._update_appearance_binding)
        tk.PanedWindow.destroy(self)

    def is_visible(self):
        if not isinstance(self.master, AutomaticPanedWindow):
            return self.winfo_ismapped()
        else:
            return self in self.master.pane_widgets()

    def pane_widgets(self):
        result = []
        for pane in self.panes():
            # pane is not the widget but some kind of reference object
            assert not isinstance(pane, tk.Widget)
            result.append(self.nametowidget(str(pane)))
        return result

    def _on_window_resize(self, event):
        if event.width < 10 or event.height < 10:
            return
        window = self.winfo_toplevel()
        window_size = (window.winfo_width(), window.winfo_height())
        initializing = hasattr(window, "initializing") and window.initializing

        if (
            not initializing
            and not self._restoring_pane_sizes
            and (window_size != self._last_window_size or self._full_size_not_final)
        ):
            self._check_restore_preferred_sizes()
            self._last_window_size = window_size

    def _on_mouse_dragged(self, event):
        if event.widget == self and not self._restoring_pane_sizes:
            self._update_preferred_sizes()

    def _update_preferred_sizes(self):
        for pane in self.pane_widgets():
            if pane.preferred_size_in_pw is not None:
                if self.cget("orient") == "horizontal":
                    current_size = pane.winfo_width()
                else:
                    current_size = pane.winfo_height()

                if current_size > 20:
                    pane.preferred_size_in_pw = current_size

                    # paneconfig width/height effectively puts
                    # unexplainable maxsize to some panes
                    # if self.cget("orient") == "horizontal":
                    #    self.paneconfig(pane, width=current_size)
                    # else:
                    #    self.paneconfig(pane, height=current_size)
                    #
            # else:
            #    self.paneconfig(pane, width=1000, height=1000)

    def _check_restore_preferred_sizes(self):
        window = self.winfo_toplevel()
        if getattr(window, "initializing", False):
            return

        try:
            self._restoring_pane_sizes = True
            self._restore_preferred_sizes()
        finally:
            self._restoring_pane_sizes = False

    def _restore_preferred_sizes(self):
        total_preferred_size = 0
        panes_without_preferred_size = []

        panes = self.pane_widgets()
        for pane in panes:
            if not hasattr(pane, "preferred_size_in_pw"):
                # child isn't fully constructed yet
                return

            if pane.preferred_size_in_pw is None:
                panes_without_preferred_size.append(pane)
                # self.paneconfig(pane, width=1000, height=1000)
            else:
                total_preferred_size += pane.preferred_size_in_pw

                # Without updating pane width/height attribute
                # the preferred size may lose effect when squeezing
                # non-preferred panes too small. Also zooming/unzooming
                # changes the supposedly fixed panes ...
                #
                # but
                # paneconfig width/height effectively puts
                # unexplainable maxsize to some panes
                # if self.cget("orient") == "horizontal":
                #    self.paneconfig(pane, width=pane.preferred_size_in_pw)
                # else:
                #    self.paneconfig(pane, height=pane.preferred_size_in_pw)

        assert len(panes_without_preferred_size) <= 1

        size = self._get_size()
        if size is None:
            return

        leftover_size = self._get_size() - total_preferred_size
        used_size = 0
        for i, pane in enumerate(panes[:-1]):
            used_size += pane.preferred_size_in_pw or leftover_size
            self._place_sash(i, used_size)
            used_size += int(str(self.cget("sashwidth")))

    def _get_size(self):
        if self.cget("orient") == tk.HORIZONTAL:
            result = self.winfo_width()
        else:
            result = self.winfo_height()

        if result < 20:
            # Not ready yet
            return None
        else:
            return result

    def _place_sash(self, i, distance):
        if self.cget("orient") == tk.HORIZONTAL:
            self.sash_place(i, distance, 0)
        else:
            self.sash_place(i, 0, distance)

    def _update_visibility(self):
        if not isinstance(self.master, AutomaticPanedWindow):
            return

        if len(self.panes()) == 0 and self.is_visible():
            self.master.forget(self)

        if len(self.panes()) > 0 and not self.is_visible():
            self.master.insert("auto", self)

    def _update_appearance(self, event=None):
        self.configure(sashwidth=lookup_style_option("Sash", "sashthickness", 10))
        self.configure(background=lookup_style_option("TPanedWindow", "background"))


class ClosableNotebook(ttk.Notebook):
    def __init__(self, master, style="ButtonNotebook.TNotebook", **kw):
        super().__init__(master, style=style, **kw)

        self.tab_menu = self.create_tab_menu()
        self._popup_index = None
        self.pressed_index = None

        self.bind("<ButtonPress-1>", self._letf_btn_press, True)
        self.bind("<ButtonRelease-1>", self._left_btn_release, True)
        if running_on_mac_os():
            self.bind("<ButtonPress-2>", self._right_btn_press, True)
            self.bind("<Control-Button-1>", self._right_btn_press, True)
        else:
            self.bind("<ButtonPress-3>", self._right_btn_press, True)

        # self._check_update_style()

    def create_tab_menu(self):
        menu = tk.Menu(
            self.winfo_toplevel(), tearoff=False, **get_style_configuration("Menu")
        )
        menu.add_command(label="Close", command=self._close_tab_from_menu)
        menu.add_command(label="Close others", command=self._close_other_tabs)
        menu.add_command(label="Close all", command=self.close_tabs)
        return menu

    def _letf_btn_press(self, event):
        try:
            elem = self.identify(event.x, event.y)
            index = self.index("@%d,%d" % (event.x, event.y))

            if "closebutton" in elem:
                self.state(["pressed"])
                self.pressed_index = index
        except Exception:
            # may fail, if clicked outside of tab
            return

    def _left_btn_release(self, event):
        if not self.instate(["pressed"]):
            return

        try:
            elem = self.identify(event.x, event.y)
            index = self.index("@%d,%d" % (event.x, event.y))
        except Exception:
            # may fail, when mouse is dragged
            return
        else:
            if "closebutton" in elem and self.pressed_index == index:
                self.close_tab(index)

            self.state(["!pressed"])
        finally:
            self.pressed_index = None

    def _right_btn_press(self, event):
        try:
            index = self.index("@%d,%d" % (event.x, event.y))
            self._popup_index = index
            self.tab_menu.tk_popup(*self.winfo_toplevel().winfo_pointerxy())
        except Exception:
            logging.exception("Opening tab menu")

    def _close_tab_from_menu(self):
        self.close_tab(self._popup_index)

    def _close_other_tabs(self):
        self.close_tabs(self._popup_index)

    def close_tabs(self, except_index=None):
        for tab_index in reversed(range(len(self.winfo_children()))):
            if except_index is not None and tab_index == except_index:
                continue
            else:
                self.close_tab(tab_index)

    def close_tab(self, index):
        child = self.get_child_by_index(index)
        if hasattr(child, "close"):
            child.close()
        else:
            self.forget(index)
            child.destroy()

    def get_child_by_index(self, index):
        tab_id = self.tabs()[index]
        if tab_id:
            return self.nametowidget(tab_id)
        else:
            return None

    def get_current_child(self):
        child_id = self.select()
        if child_id:
            return self.nametowidget(child_id)
        else:
            return None

    def focus_set(self):
        editor = self.get_current_child()
        if editor:
            editor.focus_set()
        else:
            super().focus_set()

    def _check_update_style(self):
        style = ttk.Style()
        if "closebutton" in style.element_names():
            # It's done already
            return

        # respect if required images have been defined already
        if "img_close" not in self.image_names():
            img_dir = os.path.join(os.path.dirname(__file__), "res")
            ClosableNotebook._close_img = tk.PhotoImage(
                "img_tab_close", file=os.path.join(img_dir, "tab_close.gif")
            )
            ClosableNotebook._close_active_img = tk.PhotoImage(
                "img_tab_close_active",
                file=os.path.join(img_dir, "tab_close_active.gif"),
            )

        style.element_create(
            "closebutton",
            "image",
            "img_tab_close",
            ("active", "pressed", "!disabled", "img_tab_close_active"),
            ("active", "!disabled", "img_tab_close_active"),
            border=8,
            sticky="",
        )

        style.layout(
            "ButtonNotebook.TNotebook.Tab",
            [
                (
                    "Notebook.tab",
                    {
                        "sticky": "nswe",
                        "children": [
                            (
                                "Notebook.padding",
                                {
                                    "side": "top",
                                    "sticky": "nswe",
                                    "children": [
                                        (
                                            "Notebook.focus",
                                            {
                                                "side": "top",
                                                "sticky": "nswe",
                                                "children": [
                                                    (
                                                        "Notebook.label",
                                                        {"side": "left", "sticky": ""},
                                                    ),
                                                    (
                                                        "Notebook.closebutton",
                                                        {"side": "left", "sticky": ""},
                                                    ),
                                                ],
                                            },
                                        )
                                    ],
                                },
                            )
                        ],
                    },
                )
            ],
        )


class AutomaticNotebook(ClosableNotebook):
    """
    Enables inserting views according to their position keys.
    Remember its own position key. Automatically updates its visibility.
    """

    def __init__(self, master, position_key, preferred_size_in_pw=None):
        if get_workbench().get_ui_mode() == "simple":
            style = "TNotebook"
        else:
            style = "ButtonNotebook.TNotebook"
        super().__init__(master, style=style, padding=0)
        self.position_key = position_key

        # should be in the end, so that it can be detected when
        # constructor hasn't completed yet
        self.preferred_size_in_pw = preferred_size_in_pw

    def add(self, child, **kw):
        super().add(child, **kw)
        self._update_visibility()

    def insert(self, pos, child, **kw):
        if pos == "auto":
            for sibling in map(self.nametowidget, self.tabs()):
                if (
                    not hasattr(sibling, "position_key")
                    or sibling.position_key == None
                    or sibling.position_key > child.position_key
                ):
                    pos = sibling
                    break
            else:
                pos = "end"

        super().insert(pos, child, **kw)
        self._update_visibility()

    def hide(self, tab_id):
        super().hide(tab_id)
        self._update_visibility()

    def forget(self, tab_id):
        if tab_id in self.tabs() or tab_id in self.winfo_children():
            super().forget(tab_id)
        self._update_visibility()

    def is_visible(self):
        return self in self.master.pane_widgets()

    def get_visible_child(self):
        for child in self.winfo_children():
            if str(child) == str(self.select()):
                return child

        return None

    def _update_visibility(self):
        if not isinstance(self.master, AutomaticPanedWindow):
            return
        if len(self.tabs()) == 0 and self.is_visible():
            self.master.remove(self)

        if len(self.tabs()) > 0 and not self.is_visible():
            self.master.insert("auto", self)


class TreeFrame(ttk.Frame):
    def __init__(
        self,
        master,
        columns,
        displaycolumns="#all",
        show_scrollbar=True,
        borderwidth=0,
        relief="flat",
        **tree_kw
    ):
        ttk.Frame.__init__(self, master, borderwidth=borderwidth, relief=relief)
        # http://wiki.tcl.tk/44444#pagetoc50f90d9a
        self.vert_scrollbar = ttk.Scrollbar(
            self, orient=tk.VERTICAL, style=scrollbar_style("Vertical")
        )
        if show_scrollbar:
            self.vert_scrollbar.grid(row=0, column=1, sticky=tk.NSEW)

        self.tree = ttk.Treeview(
            self,
            columns=columns,
            displaycolumns=displaycolumns,
            yscrollcommand=self.vert_scrollbar.set,
            **tree_kw
        )
        self.tree["show"] = "headings"
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.vert_scrollbar["command"] = self.tree.yview
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.tree.bind("<<TreeviewSelect>>", self.on_select, "+")
        self.tree.bind("<Double-Button-1>", self.on_double_click, "+")

    def _clear_tree(self):
        for child_id in self.tree.get_children():
            self.tree.delete(child_id)

    def clear(self):
        self._clear_tree()

    def on_select(self, event):
        pass

    def on_double_click(self, event):
        pass


def scrollbar_style(orientation):
    # In mac ttk.Scrollbar uses native rendering unless style attribute is set
    # see http://wiki.tcl.tk/44444#pagetoc50f90d9a
    # Native rendering doesn't look good in dark themes
    if running_on_mac_os() and get_workbench().uses_dark_ui_theme():
        return orientation + ".TScrollbar"
    else:
        return None


def sequence_to_accelerator(sequence):
    """Translates Tk event sequence to customary shortcut string
    for showing in the menu"""

    if not sequence:
        return ""

    if not sequence.startswith("<"):
        return sequence

    accelerator = (
        sequence.strip("<>")
        .replace("Key-", "")
        .replace("KeyPress-", "")
        .replace("Control", "Ctrl")
    )

    # Tweaking individual parts
    parts = accelerator.split("-")
    # tkinter shows shift with capital letter, but in shortcuts it's customary to include it explicitly
    if len(parts[-1]) == 1 and parts[-1].isupper() and not "Shift" in parts:
        parts.insert(-1, "Shift")

    # even when shift is not required, it's customary to show shortcut with capital letter
    if len(parts[-1]) == 1:
        parts[-1] = parts[-1].upper()

    accelerator = "+".join(parts)

    # Post processing
    accelerator = (
        accelerator.replace("Minus", "-")
        .replace("minus", "-")
        .replace("Plus", "+")
        .replace("plus", "+")
    )

    return accelerator


def get_zoomed(toplevel):
    if "-zoomed" in toplevel.wm_attributes():  # Linux
        return bool(toplevel.wm_attributes("-zoomed"))
    else:  # Win/Mac
        return toplevel.wm_state() == "zoomed"


def set_zoomed(toplevel, value):
    if "-zoomed" in toplevel.wm_attributes():  # Linux
        toplevel.wm_attributes("-zoomed", str(int(value)))
    else:  # Win/Mac
        if value:
            toplevel.wm_state("zoomed")
        else:
            toplevel.wm_state("normal")


class EnhancedTextWithLogging(tktextext.EnhancedText):
    def direct_insert(self, index, chars, tags=None, **kw):
        try:
            # try removing line numbers
            # TODO: shouldn't it take place only on paste?
            # TODO: does it occur when opening a file with line numbers in it?
            # if self._propose_remove_line_numbers and isinstance(chars, str):
            #    chars = try_remove_linenumbers(chars, self)
            concrete_index = self.index(index)
            return tktextext.EnhancedText.direct_insert(
                self, index, chars, tags=tags, **kw
            )
        finally:
            get_workbench().event_generate(
                "TextInsert",
                index=concrete_index,
                text=chars,
                tags=tags,
                text_widget=self,
            )

    def direct_delete(self, index1, index2=None, **kw):
        try:
            # index1 may be eg "sel.first" and it doesn't make sense *after* deletion
            concrete_index1 = self.index(index1)
            if index2 is not None:
                concrete_index2 = self.index(index2)
            else:
                concrete_index2 = None

            return tktextext.EnhancedText.direct_delete(
                self, index1, index2=index2, **kw
            )
        finally:
            get_workbench().event_generate(
                "TextDelete",
                index1=concrete_index1,
                index2=concrete_index2,
                text_widget=self,
            )


class SafeScrollbar(ttk.Scrollbar):
    def __init__(self, master=None, **kw):
        super().__init__(master=master, **kw)

    def set(self, first, last):
        try:
            ttk.Scrollbar.set(self, first, last)
        except Exception:
            traceback.print_exc()


class AutoScrollbar(SafeScrollbar):
    # http://effbot.org/zone/tkinter-autoscrollbar.htm
    # a vert_scrollbar that hides itself if it's not needed.  only
    # works if you use the grid geometry manager.

    def __init__(self, master=None, **kw):
        super().__init__(master=master, **kw)

    def set(self, first, last):
        if float(first) <= 0.0 and float(last) >= 1.0:
            self.grid_remove()
        elif float(first) > 0.001 or float(last) < 0.009:
            # with >0 and <1 it occasionally made scrollbar wobble back and forth
            self.grid()
        ttk.Scrollbar.set(self, first, last)

    def pack(self, **kw):
        raise tk.TclError("cannot use pack with this widget")

    def place(self, **kw):
        raise tk.TclError("cannot use place with this widget")


def update_entry_text(entry, text):
    original_state = entry.cget("state")
    entry.config(state="normal")
    entry.delete(0, "end")
    entry.insert(0, text)
    entry.config(state=original_state)


class VerticallyScrollableFrame(ttk.Frame):
    # http://tkinter.unpythonic.net/wiki/VerticalScrolledFrame

    def __init__(self, master):
        ttk.Frame.__init__(self, master)

        # set up scrolling with canvas
        vscrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        self.canvas = tk.Canvas(
            self, bd=0, highlightthickness=0, yscrollcommand=vscrollbar.set
        )
        vscrollbar.config(command=self.canvas.yview)
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)
        vscrollbar.grid(row=0, column=1, sticky=tk.NSEW)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.interior = ttk.Frame(self.canvas)
        self.interior_id = self.canvas.create_window(
            0, 0, window=self.interior, anchor=tk.NW
        )
        self.bind("<Configure>", self._configure_interior, "+")
        self.bind("<Expose>", self._expose, "+")

    def _expose(self, event):
        self.update_idletasks()
        self.update_scrollbars()

    def _configure_interior(self, event):
        self.update_scrollbars()

    def update_scrollbars(self):
        # update the scrollbars to match the size of the inner frame
        size = (self.canvas.winfo_width(), self.interior.winfo_reqheight())
        self.canvas.config(scrollregion="0 0 %s %s" % size)
        if (
            self.interior.winfo_reqwidth() != self.canvas.winfo_width()
            and self.canvas.winfo_width() > 10
        ):
            # update the interior's width to fit canvas
            # print("CAWI", self.canvas.winfo_width())
            self.canvas.itemconfigure(self.interior_id, width=self.canvas.winfo_width())


class ScrollableFrame(ttk.Frame):
    # http://tkinter.unpythonic.net/wiki/VerticalScrolledFrame

    def __init__(self, master):
        ttk.Frame.__init__(self, master)

        # set up scrolling with canvas
        vscrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        hscrollbar = ttk.Scrollbar(self, orient=tk.HORIZONTAL)
        self.canvas = tk.Canvas(
            self, bd=0, highlightthickness=0, yscrollcommand=vscrollbar.set
        )
        vscrollbar.config(command=self.canvas.yview)
        hscrollbar.config(command=self.canvas.xview)

        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)
        vscrollbar.grid(row=0, column=1, sticky=tk.NSEW)
        hscrollbar.grid(row=1, column=0, sticky=tk.NSEW)

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.interior = ttk.Frame(self.canvas)
        self.interior.columnconfigure(0, weight=1)
        self.interior.rowconfigure(0, weight=1)
        self.interior_id = self.canvas.create_window(
            0, 0, window=self.interior, anchor=tk.NW
        )
        self.bind("<Configure>", self._configure_interior, "+")
        self.bind("<Expose>", self._expose, "+")

    def _expose(self, event):
        self.update_idletasks()
        self._configure_interior(event)

    def _configure_interior(self, event):
        # update the scrollbars to match the size of the inner frame
        size = (self.canvas.winfo_reqwidth(), self.interior.winfo_reqheight())
        self.canvas.config(scrollregion="0 0 %s %s" % size)


class ThemedListbox(tk.Listbox):
    def __init__(self, master=None, cnf={}, **kw):
        super().__init__(master=master, cnf=cnf, **kw)

        self._ui_theme_change_binding = self.bind(
            "<<ThemeChanged>>", self._reload_theme_options, True
        )
        self._reload_theme_options()

    def _reload_theme_options(self, event=None):
        style = ttk.Style()

        states = []
        if self["state"] == "disabled":
            states.append("disabled")

        # Following crashes when a combobox is focused
        # if self.focus_get() == self:
        #    states.append("focus")
        opts = {}
        for key in [
            "background",
            "foreground",
            "highlightthickness",
            "highlightcolor",
            "highlightbackground",
        ]:
            value = style.lookup(self.get_style_name(), key, states)
            if value:
                opts[key] = value

        self.configure(opts)

    def get_style_name(self):
        return "Listbox"

    def destroy(self):
        self.unbind("<<ThemeChanged>>", self._ui_theme_change_binding)
        super().destroy()


class ToolTip:
    """Taken from http://www.voidspace.org.uk/python/weblog/arch_d7_2006_07_01.shtml"""

    def __init__(self, widget, options):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0
        self.options = options

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, _, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 27
        y = y + cy + self.widget.winfo_rooty() + 27
        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        if running_on_mac_os():
            # TODO: maybe it's because of Tk 8.5, not because of Mac
            tw.wm_transient(self.widget)
        tw.wm_geometry("+%d+%d" % (x, y))
        try:
            # For Mac OS
            tw.tk.call(
                "::tk::unsupported::MacWindowStyle",
                "style",
                tw._w,
                "help",
                "noActivates",
            )
        except tk.TclError:
            pass
        label = tk.Label(tw, text=self.text, **self.options)
        label.pack()

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()


def create_tooltip(widget, text, **kw):
    options = get_style_configuration("Tooltip").copy()
    options.setdefault("background", "#ffffe0")
    options.setdefault("relief", "solid")
    options.setdefault("borderwidth", 1)
    options.setdefault("padx", 1)
    options.setdefault("pady", 0)
    options.update(kw)

    toolTip = ToolTip(widget, options)

    def enter(event):
        toolTip.showtip(text)

    def leave(event):
        toolTip.hidetip()

    widget.bind("<Enter>", enter)
    widget.bind("<Leave>", leave)


class NoteBox(tk.Toplevel):
    def __init__(self, master=None, max_default_width=300, **kw):
        super().__init__(master=master, highlightthickness=0, **kw)

        self._max_default_width = max_default_width

        self.wm_overrideredirect(True)
        if running_on_mac_os():
            # TODO: maybe it's because of Tk 8.5, not because of Mac
            self.wm_transient(master)
        try:
            # For Mac OS
            self.tk.call(
                "::tk::unsupported::MacWindowStyle",
                "style",
                self._w,
                "help",
                "noActivates",
            )
        except tk.TclError:
            pass

        self._current_chars = ""
        self._click_bindings = {}

        self.padx = 5
        self.pady = 5
        self.text = TweakableText(
            self,
            background="#ffffe0",
            borderwidth=1,
            relief="solid",
            undo=False,
            read_only=True,
            font="TkDefaultFont",
            highlightthickness=0,
            padx=self.padx,
            pady=self.pady,
            wrap="word",
        )

        self.text.grid(row=0, column=0, sticky="nsew")

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.text.bind("<Escape>", self.close, True)

        # tk._default_root.bind_all("<1>", self._close_maybe, True)
        # tk._default_root.bind_all("<Key>", self.close, True)

        self.withdraw()

    def clear(self):
        for tag in self._click_bindings:
            self.text.tag_unbind(tag, "<1>", self._click_bindings[tag])
            self.text.tag_remove(tag, "1.0", "end")

        self.text.direct_delete("1.0", "end")
        self._current_chars = ""
        self._click_bindings.clear()

    def set_content(self, *items):
        self.clear()

        for item in items:
            if isinstance(item, str):
                self.text.direct_insert("1.0", item)
                self._current_chars = item
            else:
                assert isinstance(item, (list, tuple))
                chars, *props = item
                if len(props) > 0 and callable(props[-1]):
                    tags = tuple(props[:-1])
                    click_handler = props[-1]
                else:
                    tags = tuple(props)
                    click_handler = None

                self.append_text(chars, tags, click_handler)

            self.text.see("1.0")

    def append_text(self, chars, tags=(), click_handler=None):
        tags = tuple(tags)

        if click_handler is not None:
            click_tag = "click_%d" % len(self._click_bindings)
            tags = tags + (click_tag,)
            binding = self.text.tag_bind(click_tag, "<1>", click_handler, True)
            self._click_bindings[click_tag] = binding

        self.text.direct_insert("end", chars, tags)
        self._current_chars += chars

    def place(self, target, focus=None):

        # Compute the area that will be described by this Note
        focus_x = target.winfo_rootx()
        focus_y = target.winfo_rooty()
        focus_height = target.winfo_height()

        if isinstance(focus, TextRange):
            assert isinstance(target, tk.Text)
            topleft = target.bbox("%d.%d" % (focus.lineno, focus.col_offset))
            if focus.end_col_offset == 0:
                botright = target.bbox(
                    "%d.%d lineend" % (focus.end_lineno - 1, focus.end_lineno - 1)
                )
            else:
                botright = target.bbox(
                    "%d.%d" % (focus.end_lineno, focus.end_col_offset)
                )

            if topleft and botright:
                focus_x += topleft[0]
                focus_y += topleft[1]
                focus_height = botright[1] - topleft[1] + botright[3]

        elif isinstance(focus, (list, tuple)):
            focus_x += focus[0]
            focus_y += focus[1]
            focus_height = focus[3]

        elif focus is None:
            pass

        else:
            raise TypeError("Unsupported focus")

        # Compute dimensions of the note
        font = self.text["font"]
        if isinstance(font, str):
            font = tk.font.nametofont(font)

        lines = self._current_chars.splitlines()
        max_line_width = 0
        for line in lines:

            max_line_width = max(max_line_width, font.measure(line))

        width = min(max_line_width, self._max_default_width) + self.padx * 2 + 2
        self.wm_geometry("%dx%d+%d+%d" % (width, 100, focus_x, focus_y + focus_height))

        self.update_idletasks()
        line_count = int(float(self.text.index("end")))
        line_height = font.metrics()["linespace"]

        self.wm_geometry(
            "%dx%d+%d+%d"
            % (width, line_count * line_height, focus_x, focus_y + focus_height)
        )

        # TODO: detect the situation when note doesn't fit under
        # the focus box and should be placed above

        self.deiconify()

    def show_note(
        self, *content_items: Union[str, List], target=None, focus=None
    ) -> None:

        self.set_content(*content_items)
        self.place(target, focus)

    def _close_maybe(self, event):
        if event.widget not in [self, self.text]:
            self.close(event)

    def close(self, event=None):
        self.withdraw()


def get_widget_offset_from_toplevel(widget):
    x = 0
    y = 0
    toplevel = widget.winfo_toplevel()
    while widget != toplevel:
        x += widget.winfo_x()
        y += widget.winfo_y()
        widget = widget.master
    return x, y


def create_string_var(value, modification_listener=None):
    """Creates a tk.StringVar with "modified" attribute
    showing whether the variable has been modified after creation"""
    return _create_var(tk.StringVar, value, modification_listener)


def create_int_var(value, modification_listener=None):
    """See create_string_var"""
    return _create_var(tk.IntVar, value, modification_listener)


def create_double_var(value, modification_listener=None):
    """See create_string_var"""
    return _create_var(tk.DoubleVar, value, modification_listener)


def create_boolean_var(value, modification_listener=None):
    """See create_string_var"""
    return _create_var(tk.BooleanVar, value, modification_listener)


def _create_var(class_, value, modification_listener):
    var = class_(value=value)
    var.modified = False

    def on_write(*args):
        var.modified = True
        if modification_listener:
            try:
                modification_listener()
            except Exception:
                # Otherwise whole process will be brought down
                # because for some reason Tk tries to call non-existing method
                # on variable
                get_workbench().report_exception()

    # TODO: https://bugs.python.org/issue22115 (deprecation warning)
    var.trace("w", on_write)
    return var


def shift_is_pressed(event_state):
    # http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/event-handlers.html
    # http://stackoverflow.com/q/32426250/261181
    return event_state & 0x0001


def control_is_pressed(event_state):
    # http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/event-handlers.html
    # http://stackoverflow.com/q/32426250/261181
    return event_state & 0x0004


def sequence_to_event_state_and_keycode(sequence: str) -> Optional[Tuple[int, int]]:
    # remember handlers for certain shortcuts which require
    # different treatment on non-latin keyboards
    if sequence[0] != "<":
        return None

    parts = sequence.strip("<").strip(">").split("-")
    # support only latin letters for now
    if parts[-1].lower() not in list("abcdefghijklmnopqrstuvwxyz"):
        return None

    letter = parts.pop(-1)
    if "Key" in parts:
        parts.remove("Key")
    if "key" in parts:
        parts.remove("key")

    modifiers = {part.lower() for part in parts}

    if letter.isupper():
        modifiers.add("shift")

    if modifiers not in [{"control"}, {"control", "shift"}]:
        # don't support others for now
        return None

    event_state = 0
    # http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/event-handlers.html
    # https://stackoverflow.com/questions/32426250/python-documentation-and-or-lack-thereof-e-g-keyboard-event-state
    for modifier in modifiers:
        if modifier == "shift":
            event_state |= 0x0001
        elif modifier == "control":
            event_state |= 0x0004
        else:
            # unsupported modifier
            return None

    # for latin letters keycode is same as its ascii code
    return (event_state, ord(letter.upper()))


def select_sequence(win_version, mac_version, linux_version=None):
    if running_on_windows():
        return win_version
    elif running_on_mac_os():
        return mac_version
    elif running_on_linux() and linux_version:
        return linux_version
    else:
        return win_version


def try_remove_linenumbers(text, master):
    try:
        if has_line_numbers(text) and messagebox.askyesno(
            title="Remove linenumbers",
            message="Do you want to remove linenumbers from pasted text?",
            default=messagebox.YES,
            master=master,
        ):
            return remove_line_numbers(text)
        else:
            return text
    except Exception:
        traceback.print_exc()
        return text


def has_line_numbers(text):
    lines = text.splitlines()
    return len(lines) > 2 and all(
        [len(split_after_line_number(line)) == 2 for line in lines]
    )


def split_after_line_number(s):
    parts = re.split(r"(^\s*\d+\.?)", s)
    if len(parts) == 1:
        return parts
    else:
        assert len(parts) == 3 and parts[0] == ""
        return parts[1:]


def remove_line_numbers(s):
    cleaned_lines = []
    for line in s.splitlines():
        parts = split_after_line_number(line)
        if len(parts) != 2:
            return s
        else:
            cleaned_lines.append(parts[1])

    return textwrap.dedent(("\n".join(cleaned_lines)) + "\n")


def center_window(win, master=None):
    # looks like it doesn't take window border into account
    win.update_idletasks()

    if getattr(master, "initializing", False):
        # can't get reliable positions when main window is not in mainloop yet
        left = (win.winfo_screenwidth() - 600) // 2
        top = (win.winfo_screenheight() - 400) // 2
    else:
        if master is None:
            left = win.winfo_screenwidth() - win.winfo_width() // 2
            top = win.winfo_screenheight() - win.winfo_height() // 2
        else:
            left = (
                master.winfo_rootx()
                + master.winfo_width() // 2
                - win.winfo_width() // 2
            )
            top = (
                master.winfo_rooty()
                + master.winfo_height() // 2
                - win.winfo_height() // 2
            )

    win.geometry("+%d+%d" % (left, top))


class WaitingDialog(tk.Toplevel):
    def __init__(
        self, master, async_result, description, title="Please wait!", timeout=None
    ):
        self._async_result = async_result
        super().__init__(master)
        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        self.title(title)
        self.resizable(height=tk.FALSE, width=tk.FALSE)
        # self.protocol("WM_DELETE_WINDOW", self._close)
        self.desc_label = ttk.Label(self, text=description, wraplength=300)
        self.desc_label.grid(padx=20, pady=20)

        self.update_idletasks()

        self.timeout = timeout
        self.start_time = time.time()
        self.after(500, self._poll)

    def _poll(self):
        if self._async_result.ready():
            self._close()
        elif self.timeout and time.time() - self.start_time > self.timeout:
            raise TimeoutError()
        else:
            self.after(500, self._poll)
            self.desc_label["text"] = self.desc_label["text"] + "."

    def _close(self):
        self.destroy()


def run_with_waiting_dialog(master, action, args=(), description="Working"):
    # http://stackoverflow.com/a/14299004/261181
    from multiprocessing.pool import ThreadPool

    pool = ThreadPool(processes=1)

    async_result = pool.apply_async(action, args)
    dlg = WaitingDialog(master, async_result, description=description)
    show_dialog(dlg, master)

    return async_result.get()


class FileCopyDialog(tk.Toplevel):
    def __init__(self, master, source, destination, description=None, fsync=True):
        self._source = source
        self._destination = destination
        self._old_bytes_copied = 0
        self._bytes_copied = 0
        self._fsync = fsync
        self._done = False
        self._cancelled = False
        self._closed = False

        super().__init__(master)

        main_frame = ttk.Frame(self)  # To get styled background
        main_frame.grid(row=0, column=0, sticky="nsew")
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.title("Copying")

        if description is None:
            description = "Copying\n  %s\nto\n  %s" % (source, destination)

        label = ttk.Label(main_frame, text=description)
        label.grid(row=0, column=0, columnspan=2, sticky="nw", padx=15, pady=15)

        self._bar = ttk.Progressbar(
            main_frame, maximum=os.path.getsize(source), length=200
        )
        self._bar.grid(row=1, column=0, columnspan=2, sticky="nsew", padx=15, pady=0)

        self._cancel_button = ttk.Button(
            main_frame, text="Cancel", command=self._cancel
        )
        self._cancel_button.grid(row=2, column=1, sticky="ne", padx=15, pady=15)
        self._bar.focus_set()

        main_frame.columnconfigure(0, weight=1)

        self._update_progress()

        self.bind(
            "<Escape>", self._cancel, True
        )  # escape-close only if process has completed
        self.protocol("WM_DELETE_WINDOW", self._cancel)
        center_window(self, master)
        self._start()

    def _start(self):
        def work():
            self._copy_progess = 0

            with open(self._source, "rb") as fsrc:
                with open(self._destination, "wb") as fdst:
                    while True:
                        buf = fsrc.read(16 * 1024)
                        if not buf:
                            break

                        fdst.write(buf)
                        fdst.flush()
                        if self._fsync:
                            os.fsync(fdst)
                        self._bytes_copied += len(buf)

            self._done = True

        threading.Thread(target=work, daemon=True).start()

    def _update_progress(self):
        if self._done:
            if not self._closed:
                self._close()
            return

        self._bar.step(self._bytes_copied - self._old_bytes_copied)
        self._old_bytes_copied = self._bytes_copied

        self.after(100, self._update_progress)

    def _close(self):
        self.destroy()
        self._closed = True

    def _cancel(self, event=None):
        self._cancelled = True
        self._close()


class ChoiceDialog(tk.Toplevel):
    def __init__(
        self, master=None, title="Choose one", question: str = "Choose one:", choices=[]
    ) -> None:
        super().__init__(master=master)

        self.title(title)
        self.resizable(False, False)

        self.columnconfigure(0, weight=1)

        row = 0
        question_label = ttk.Label(self, text=question)
        question_label.grid(
            row=row, column=0, columnspan=2, sticky="w", padx=20, pady=20
        )
        row += 1

        self.var = tk.StringVar()
        for choice in choices:
            rb = ttk.Radiobutton(self, text=choice, variable=self.var, value=choice)
            rb.grid(row=row, column=0, columnspan=2, sticky="w", padx=20)
            row += 1

        ok_button = ttk.Button(self, text="OK", command=self._ok, default="active")
        ok_button.grid(row=row, column=0, sticky="e", pady=20)

        cancel_button = ttk.Button(self, text="Cancel", command=self._cancel)
        cancel_button.grid(row=row, column=1, sticky="e", padx=20, pady=20)

        self.bind("<Escape>", self._cancel, True)
        self.bind("<Return>", self._ok, True)
        self.protocol("WM_DELETE_WINDOW", self._cancel)

        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")

        center_window(self, master)

    def _ok(self):
        self.result = self.var.get()
        if not self.result:
            self.result = None

        self.destroy()

    def _cancel(self):
        self.result = None
        self.destroy()


def ask_one_from_choices(
    master=None, title="Choose one", question: str = "Choose one:", choices=[]
):
    dlg = ChoiceDialog(master, title, question, choices)
    show_dialog(dlg, master)
    return dlg.result


class SubprocessDialog(tk.Toplevel):
    """Shows incrementally the output of given subprocess.
    Allows cancelling"""

    def __init__(
        self,
        master,
        proc,
        title,
        long_description=None,
        autoclose=True,
        conclusion="Done.",
    ):
        self._closed = False
        self._proc = proc
        self.stdout = ""
        self.stderr = ""
        self._stdout_thread = None
        self._stderr_thread = None
        self.returncode = None
        self.cancelled = False
        self._autoclose = autoclose
        self._event_queue = collections.deque()
        self._conclusion = conclusion

        tk.Toplevel.__init__(self, master)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        main_frame = ttk.Frame(self)  # To get styled background
        main_frame.grid(sticky="nsew")

        text_font = tk.font.nametofont("TkFixedFont").copy()
        text_font["size"] = int(text_font["size"] * 0.9)
        text_font["family"] = "Courier" if running_on_mac_os() else "Courier New"
        text_frame = tktextext.TextFrame(
            main_frame,
            read_only=True,
            horizontal_scrollbar=False,
            background=lookup_style_option("TFrame", "background"),
            font=text_font,
            wrap="word",
        )
        text_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=15, pady=15)
        self.text = text_frame.text
        self.text["width"] = 60
        self.text["height"] = 7
        if long_description is not None:
            self.text.direct_insert("1.0", long_description + "\n\n")

        self.button = ttk.Button(main_frame, text="Cancel", command=self._close)
        self.button.grid(row=1, column=0, pady=(0, 15))

        main_frame.rowconfigure(0, weight=1)
        main_frame.columnconfigure(0, weight=1)

        self.title(title)
        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        # self.resizable(height=tk.FALSE, width=tk.FALSE)
        self.text.focus_set()

        self.bind(
            "<Escape>", self._close_if_done, True
        )  # escape-close only if process has completed
        self.protocol("WM_DELETE_WINDOW", self._close)
        center_window(self, master)

        self._start_listening()

    def _start_listening(self):
        def listen_stream(stream_name):
            stream = getattr(self._proc, stream_name)
            while True:
                data = stream.readline()
                self._event_queue.append((stream_name, data))
                setattr(self, stream_name, getattr(self, stream_name) + data)
                if data == "":
                    break

            self.returncode = self._proc.wait()

        self._stdout_thread = threading.Thread(
            target=listen_stream, args=["stdout"], daemon=True
        )
        self._stdout_thread.start()
        if self._proc.stderr is not None:
            self._stderr_thread = threading.Thread(
                target=listen_stream, args=["stderr"], daemon=True
            )
            self._stderr_thread.start()

        def poll_output_events():
            if self._closed:
                return

            while len(self._event_queue) > 0:
                stream_name, data = self._event_queue.popleft()
                self.text.direct_insert("end", data, tags=(stream_name,))
                self.text.see("end")

            self.returncode = self._proc.poll()
            if self.returncode == None:
                self.after(200, poll_output_events)
            else:
                self.button["text"] = "OK"
                self.button.focus_set()
                if self.returncode != 0:
                    self.text.direct_insert("end", "\n\nReturn code: ", ("stderr",))
                elif self._autoclose:
                    self._close()
                else:
                    self.text.direct_insert("end", "\n\n" + self._conclusion)
                    self.text.see("end")

        poll_output_events()

    def _close_if_done(self, event):
        if self._proc.poll() is not None:
            self._close(event)

    def _close(self, event=None):
        if self._proc.poll() is None:
            if messagebox.askyesno(
                "Cancel the process?",
                "The process is still running.\nAre you sure you want to cancel?",
            ):
                # try gently first
                try:
                    if running_on_windows():
                        os.kill(
                            self._proc.pid, signal.CTRL_BREAK_EVENT
                        )  # @UndefinedVariable
                    else:
                        os.kill(self._proc.pid, signal.SIGINT)

                    self._proc.wait(2)
                except subprocess.TimeoutExpired:
                    if self._proc.poll() is None:
                        # now let's be more concrete
                        self._proc.kill()

                self.cancelled = True
                # Wait for threads to finish
                self._stdout_thread.join(2)
                if self._stderr_thread is not None:
                    self._stderr_thread.join(2)

                # fetch output about cancelling
                while len(self._event_queue) > 0:
                    stream_name, data = self._event_queue.popleft()
                    self.text.direct_insert("end", data, tags=(stream_name,))
                self.text.direct_insert("end", "\n\nPROCESS CANCELLED")
                self.text.see("end")

            else:
                return
        else:
            self._closed = True
            self.destroy()


def get_busy_cursor():
    if running_on_windows():
        return "wait"
    elif running_on_mac_os():
        return "spinning"
    else:
        return "watch"


def get_tk_version_str():
    return tk._default_root.tk.call("info", "patchlevel")


def get_tk_version_info():
    result = []
    for part in get_tk_version_str().split("."):
        try:
            result.append(int(part))
        except Exception:
            result.append(0)
    return tuple(result)


def get_style_configuration(style_name, default={}):
    style = ttk.Style()
    # NB! style.configure seems to reuse the returned dict
    # Don't change it without copying first
    result = style.configure(style_name)
    if result is None:
        return default
    else:
        return result


def lookup_style_option(style_name, option_name, default=None):
    style = ttk.Style()
    setting = style.lookup(style_name, option_name)
    if setting in [None, ""]:
        return default
    elif setting == "True":
        return True
    elif setting == "False":
        return False
    else:
        return setting


def scale(value):
    return get_workbench().scale(value)


def open_path_in_system_file_manager(path):
    if running_on_mac_os():
        # http://stackoverflow.com/a/3520693/261181
        # -R doesn't allow showing hidden folders
        subprocess.Popen(["open", path])
    elif running_on_linux():
        subprocess.Popen(["xdg-open", path])
    else:
        assert running_on_windows()
        subprocess.Popen(["explorer", path])


def _get_dialog_provider():
    if platform.system() != "Linux":
        return filedialog

    if shutil.which("zenity"):
        return _ZenityDialogProvider

    # fallback
    return filedialog


def asksaveasfilename(**options):
    # https://tcl.tk/man/tcl8.6/TkCmd/getSaveFile.htm
    return _get_dialog_provider().asksaveasfilename(**options)


def askopenfilename(**options):
    # https://tcl.tk/man/tcl8.6/TkCmd/getOpenFile.htm
    return _get_dialog_provider().askopenfilename(**options)


def askopenfilenames(**options):
    # https://tcl.tk/man/tcl8.6/TkCmd/getOpenFile.htm
    return _get_dialog_provider().askopenfilenames(**options)


def askdirectory(**options):
    # https://tcl.tk/man/tcl8.6/TkCmd/chooseDirectory.htm
    return _get_dialog_provider().askdirectory(**options)


class _ZenityDialogProvider:
    # https://www.writebash.com/bash-gui/zenity-create-file-selection-dialog-224.html
    # http://linux.byexamples.com/archives/259/a-complete-zenity-dialog-examples-1/
    # http://linux.byexamples.com/archives/265/a-complete-zenity-dialog-examples-2/

    # another possibility is to use PyGobject: https://github.com/poulp/zenipy

    @classmethod
    def askopenfilename(cls, **options):
        args = cls._convert_common_options("Open file", **options)
        return cls._call(args)

    @classmethod
    def askopenfilenames(cls, **options):
        args = cls._convert_common_options("Open files", **options)
        return cls._call(args + ["--multiple"]).split("|")

    @classmethod
    def asksaveasfilename(cls, **options):
        args = cls._convert_common_options("Save as", **options)
        args.append("--save")
        if options.get("confirmoverwrite", True):
            args.append("--confirm-overwrite")

        filename = cls._call(args)
        if not filename:
            return None

        if "defaultextension" in options and "." not in os.path.basename(filename):
            filename += options["defaultextension"]

        return filename

    @classmethod
    def askdirectory(cls, **options):
        args = cls._convert_common_options("Select directory", **options)
        args.append("--directory")
        return cls._call(args)

    @classmethod
    def _convert_common_options(cls, default_title, **options):
        args = ["--file-selection", "--title=%s" % options.get("title", default_title)]

        filename = _options_to_zenity_filename(options)
        if filename:
            args.append("--filename=%s" % filename)

        parent = options.get("parent", options.get("master", None))
        if parent is not None:
            args.append("--modal")
            args.append("--attach=%s" % parent.winfo_id())

        for desc, pattern in options.get("filetypes", ()):
            # zenity requires star before extension
            pattern = pattern.replace(" .", " *.")
            if pattern.startswith("."):
                pattern = "*" + pattern

            args.append("--file-filter=%s | %s" % (desc, pattern))

        return args

    @classmethod
    def _call(cls, args):
        args = ["zenity", "--name=Thonny", "--class=Thonny"] + args
        result = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        if result.returncode == 0:
            return result.stdout.strip()
        else:
            # could check stderr, but it may contain irrelevant warnings
            return None


def _options_to_zenity_filename(options):
    if "initialdir" in options:
        if "initialfile" in options:
            return os.path.join(options["initialdir"], options["initialfile"])
        else:
            return options["initialdir"] + os.path.sep

    return None


def register_latin_shortcut(
    registry, sequence: str, handler: Callable, tester: Optional[Callable]
) -> None:
    res = sequence_to_event_state_and_keycode(sequence)
    if res is not None:
        if res not in registry:
            registry[res] = []
        registry[res].append((handler, tester))


def handle_mistreated_latin_shortcuts(registry, event):
    # tries to handle Ctrl+LatinLetter shortcuts
    # given from non-Latin keyboards
    # See: https://bitbucket.org/plas/thonny/issues/422/edit-keyboard-shortcuts-ctrl-c-ctrl-v-etc

    # only consider events with Control held
    if not event.state & 0x04:
        return

    if running_on_mac_os():
        return

    # consider only part of the state,
    # because at least on Windows, Ctrl-shortcuts' state
    # has something extra
    simplified_state = 0x04
    if shift_is_pressed(event.state):
        simplified_state |= 0x01

    # print(simplified_state, event.keycode)
    if (simplified_state, event.keycode) in registry:
        if event.keycode != ord(event.char):
            # keycode and char doesn't match,
            # this means non-latin keyboard
            for handler, tester in registry[(simplified_state, event.keycode)]:
                if tester is None or tester():
                    handler()


def show_dialog(dlg, master=None):
    if master is None:
        master = tk._default_root

    focused_widget = master.focus_get()
    center_window(dlg, master)
    dlg.transient(master)
    dlg.grab_set()
    dlg.lift()
    dlg.focus_set()
    master.wait_window(dlg)
    dlg.grab_release()
    master.lift()
    master.focus_force()
    if focused_widget is not None:
        focused_widget.focus_set()


def popen_with_ui_thread_callback(
    *Popen_args, on_completion, poll_delay=0.1, **Popen_kwargs
):
    if "encoding" not in Popen_kwargs:
        if "env" not in Popen_kwargs: 
            Popen_kwargs["env"] = os.environ.copy()
        Popen_kwargs["env"]["PYTHONIOENCODING"] = "utf-8"
        if sys.version_info >= (3,6): 
            Popen_kwargs["encoding"] = "utf-8"
    
    proc = subprocess.Popen(*Popen_args, **Popen_kwargs)
    
    # Need to read in thread in order to avoid blocking because
    # of full pipe buffer (see https://bugs.python.org/issue1256)
    out_lines = []
    err_lines = []
    
    def read_stream(stream, target_list):
        while True:
            line = stream.readline()
            if line:
                target_list.append(line)
            else:
                break
            
    t_out = threading.Thread(target=read_stream, daemon=True,
                             args=(proc.stdout, out_lines))
    t_err = threading.Thread(target=read_stream, daemon=True,
                             args=(proc.stderr, err_lines))
    t_out.start()
    t_err.start()
    
    def poll():
        if proc.poll() is not None:
            t_out.join(3)
            t_err.join(3)
            on_completion(proc, out_lines, err_lines)
            return

        tk._default_root.after(int(poll_delay * 1000), poll)

    poll()
    return proc


if __name__ == "__main__":
    root = tk.Tk()
    closa = ClosableNotebook(root)
    closa.add(ttk.Button(closa, text="B1"), text="B1")
    closa.add(ttk.Button(closa, text="B2"), text="B2")
    closa.grid()
    root.mainloop()



# -*- coding: utf-8 -*-

import os.path
import platform
import shlex
import shutil
import subprocess
import sys
import time
from typing import Optional, Sequence, Tuple

from thonny import get_workbench


def delete_dir_try_hard(path: str, hardness: int = 5) -> None:
    # Deleting the folder on Windows is not so easy task
    # http://bugs.python.org/issue15496
    for i in range(hardness):
        if os.path.exists(path):
            time.sleep(i * 0.5)
            shutil.rmtree(path, True)
        else:
            break

    if os.path.exists(path):
        # try once more but now without ignoring errors
        shutil.rmtree(path, False)


def running_on_windows() -> bool:
    return platform.system() == "Windows"


def running_on_mac_os() -> bool:
    return platform.system() == "Darwin"


def running_on_linux() -> bool:
    return platform.system() == "Linux"


def is_hidden_or_system_file(path: str) -> bool:
    if os.path.basename(path).startswith("."):
        return True
    elif running_on_windows():
        from ctypes import windll

        FILE_ATTRIBUTE_HIDDEN = 0x2
        FILE_ATTRIBUTE_SYSTEM = 0x4
        return bool(
            windll.kernel32.GetFileAttributesW(path)  # @UndefinedVariable
            & (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM)
        )
    else:
        return False


def get_win_drives() -> Sequence[str]:
    # http://stackoverflow.com/a/2288225/261181
    # http://msdn.microsoft.com/en-us/library/windows/desktop/aa364939%28v=vs.85%29.aspx
    import string
    from ctypes import windll

    all_drive_types = [
        "DRIVE_UNKNOWN",
        "DRIVE_NO_ROOT_DIR",
        "DRIVE_REMOVABLE",
        "DRIVE_FIXED",
        "DRIVE_REMOTE",
        "DRIVE_CDROM",
        "DRIVE_RAMDISK",
    ]

    required_drive_types = [
        "DRIVE_REMOVABLE",
        "DRIVE_FIXED",
        "DRIVE_REMOTE",
        "DRIVE_RAMDISK",
    ]

    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()  # @UndefinedVariable
    for letter in string.ascii_uppercase:
        drive_type = all_drive_types[
            windll.kernel32.GetDriveTypeW("%s:\\" % letter)
        ]  # @UndefinedVariable
        if bitmask & 1 and drive_type in required_drive_types:
            drives.append(letter + ":\\")
        bitmask >>= 1

    return drives


def list_volumes() -> Sequence[str]:
    "Adapted from https://github.com/ntoll/uflash/blob/master/uflash.py"
    if os.name == "posix":
        # 'posix' means we're on Linux or OSX (Mac).
        # Call the unix "mount" command to list the mounted volumes.
        mount_output = subprocess.check_output("mount").splitlines()
        return [x.split()[2].decode("utf-8") for x in mount_output]

    elif os.name == "nt":
        # 'nt' means we're on Windows.
        import ctypes

        #
        # In certain circumstances, volumes are allocated to USB
        # storage devices which cause a Windows popup to raise if their
        # volume contains no media. Wrapping the check in SetErrorMode
        # with SEM_FAILCRITICALERRORS (1) prevents this popup.
        #
        old_mode = ctypes.windll.kernel32.SetErrorMode(1)  # @UndefinedVariable
        try:
            volumes = []
            for disk in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                path = "{}:\\".format(disk)
                if os.path.exists(path):
                    volumes.append(path)

            return volumes
        finally:
            ctypes.windll.kernel32.SetErrorMode(old_mode)  # @UndefinedVariable
    else:
        # No support for unknown operating systems.
        raise NotImplementedError('OS "{}" not supported.'.format(os.name))


def get_win_volume_name(path: str) -> str:
    """
    Each disk or external device connected to windows has an attribute
    called "volume name". This function returns the volume name for
    the given disk/device.
    Code from http://stackoverflow.com/a/12056414
    """
    import ctypes

    vol_name_buf = ctypes.create_unicode_buffer(1024)
    ctypes.windll.kernel32.GetVolumeInformationW(  # @UndefinedVariable
        ctypes.c_wchar_p(path),
        vol_name_buf,
        ctypes.sizeof(vol_name_buf),
        None,
        None,
        None,
        None,
        0,
    )
    assert isinstance(vol_name_buf.value, str)
    return vol_name_buf.value


def find_volumes_by_name(volume_name: str) -> Sequence[str]:
    volumes = list_volumes()
    if os.name == "nt":
        return [
            volume
            for volume in volumes
            if get_win_volume_name(volume).upper() == volume_name.upper()
        ]
    else:
        return [volume for volume in volumes if volume.endswith(volume_name)]


def find_volume_by_name(
    volume_name: str,
    not_found_msg: str = "Could not find disk '%s'. Do you want to locate it yourself?",
    found_several_msg: str = "Found several '%s' disks. Do you want to choose one yourself?",
) -> Optional[str]:

    volumes = find_volumes_by_name(volume_name)
    if len(volumes) == 1:
        return volumes[0]
    else:
        if len(volumes) == 0:
            msg = not_found_msg % volume_name
        else:
            msg = found_several_msg % volume_name

        from tkinter.messagebox import askyesno
        from thonny.ui_utils import askdirectory

        if askyesno("Can't find suitable disk", msg):
            path = askdirectory(master=get_workbench())
            if path:
                return path

    return None


def shorten_repr(original_repr: str, max_len: int = 1000) -> str:
    if len(original_repr) > max_len:
        return original_repr[: max_len - 1] + "…"
    else:
        return original_repr


def get_python_version_string(version_info: Optional[Tuple] = None) -> str:
    if version_info is None:
        version_info = sys.version_info

    result = ".".join(map(str, version_info[:3]))
    if version_info[3] != "final":
        result += "-" + version_info[3]

    result += " (" + ("64" if sys.maxsize > 2 ** 32 else "32") + " bit)\n"

    return result


def _win_get_used_memory():
    # http://code.activestate.com/recipes/578513-get-memory-usage-of-windows-processes-using-getpro/
    import ctypes
    from ctypes import wintypes

    GetCurrentProcess = ctypes.windll.kernel32.GetCurrentProcess
    GetCurrentProcess.argtypes = []
    GetCurrentProcess.restype = wintypes.HANDLE

    SIZE_T = ctypes.c_size_t

    class PROCESS_MEMORY_COUNTERS_EX(ctypes.Structure):
        _fields_ = [
            ("cb", wintypes.DWORD),
            ("PageFaultCount", wintypes.DWORD),
            ("PeakWorkingSetSize", SIZE_T),
            ("WorkingSetSize", SIZE_T),
            ("QuotaPeakPagedPoolUsage", SIZE_T),
            ("QuotaPagedPoolUsage", SIZE_T),
            ("QuotaPeakNonPagedPoolUsage", SIZE_T),
            ("QuotaNonPagedPoolUsage", SIZE_T),
            ("PagefileUsage", SIZE_T),
            ("PeakPagefileUsage", SIZE_T),
            ("PrivateUsage", SIZE_T),
        ]

    GetProcessMemoryInfo = ctypes.windll.psapi.GetProcessMemoryInfo
    GetProcessMemoryInfo.argtypes = [
        wintypes.HANDLE,
        ctypes.POINTER(PROCESS_MEMORY_COUNTERS_EX),
        wintypes.DWORD,
    ]
    GetProcessMemoryInfo.restype = wintypes.BOOL

    def get_current_process():
        """Return handle to current process."""
        return GetCurrentProcess()

    def get_memory_info(process=None):
        """Return Win32 process memory counters structure as a dict."""
        if process is None:
            process = get_current_process()
        counters = PROCESS_MEMORY_COUNTERS_EX()
        ret = GetProcessMemoryInfo(
            process, ctypes.byref(counters), ctypes.sizeof(counters)
        )
        if not ret:
            raise ctypes.WinError()
        info = dict((name, getattr(counters, name)) for name, _ in counters._fields_)
        return info

    return get_memory_info()["PrivateUsage"]


def _unix_get_used_memory():
    # http://fa.bianp.net/blog/2013/different-ways-to-get-memory-consumption-or-lessons-learned-from-memory_profiler/
    "TODO:"


def construct_cmd_line(parts):
    return " ".join(map(shlex.quote, parts))


def parse_cmd_line(s):
    return shlex.split(s, posix=True)


def levenshtein_distance(s1, s2):
    # https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
    if len(s1) < len(s2):
        return levenshtein_distance(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = (
                previous_row[j + 1] + 1
            )  # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1  # than s2
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def levenshtein_damerau_distance(s1, s2, maxDistance):
    # https://gist.github.com/giststhebearbear/4145811
    #  get smallest string so our rows are minimized
    s1, s2 = (s1, s2) if len(s1) <= len(s2) else (s2, s1)
    #  set lengths
    l1, l2 = len(s1), len(s2)

    #  We are simulatng an NM matrix where n is the longer string
    #  and m is the shorter string. By doing this we can minimize
    #  memory usage to O(M).
    #  Since we are simulating the matrix we only maintain two rows
    #  at a time the current row and the previous rows.
    #  A move from the current cell looking at the cell before it indicates
    #  consideration of an insert operation.
    #  A move from the current cell looking at the cell above it indicates
    #  consideration of a deletion
    #  Both operations are cost 1
    #  A move from the current cell to the cell up and to the left indicates
    #  an edit operation of 0 cost for a matching character and a 1 cost for
    #  a non matching characters
    #  no row has been previously computed yet, set empty row
    #  Since this is also a Damerou-Levenshtein calculation transposition
    #  costs will be taken into account. These look back 2 characters to
    #  determine optimal cost based on a possible transposition
    #  example: aei -> aie with levensthein has a cost of 2
    #  match a, change e->i change i->e => aie
    #  Damarau-Levenshtein has a cost of 1
    #  match a, transpose ei to ie => aie
    transpositionRow = None
    prevRow = None

    #  build first leven matrix row
    #  The first row represents transformation from an empty string
    #  to the shorter string making it static [0-n]
    #  since this row is static we can set it as
    #  curRow and start computation at the second row or index 1
    curRow = [x for x in range(0, l1 + 1)]

    # use second length to loop through all the rows being built
    # we start at row one
    for rowNum in range(1, l2 + 1):
        #  set transposition, previous, and current
        #  because the rowNum always increments by one
        #  we can use rowNum to set the value representing
        #  the first column which is indicitive of transforming TO
        #  the empty string from our longer string
        #  transposition row maintains an extra row so that it is possible
        #  for us to apply Damarou's formula
        transpositionRow, prevRow, curRow = prevRow, curRow, [rowNum] + [0] * l1

        #  consider if we have passed the max distance if all paths through
        #  the transposition row are larger than the max we can stop calculating
        #  distance and return the last element in that row and return the max
        if transpositionRow:
            if not any(cellValue < maxDistance for cellValue in transpositionRow):
                return maxDistance

        for colNum in range(1, l1 + 1):
            insertionCost = curRow[colNum - 1] + 1
            deletionCost = prevRow[colNum] + 1
            changeCost = prevRow[colNum - 1] + (
                0 if s1[colNum - 1] == s2[rowNum - 1] else 1
            )
            #  set the cell value - min distance to reach this
            #  position
            curRow[colNum] = min(insertionCost, deletionCost, changeCost)

            #  test for a possible transposition optimization
            #  check to see if we have at least 2 characters
            if 1 < rowNum <= colNum:
                #  test for possible transposition
                if (
                    s1[colNum - 1] == s2[colNum - 2]
                    and s2[colNum - 1] == s1[colNum - 2]
                ):
                    curRow[colNum] = min(
                        curRow[colNum], transpositionRow[colNum - 2] + 1
                    )

    #  the last cell of the matrix is ALWAYS the shortest distance between the two strings
    return curRow[-1]


def get_file_creation_date(path_to_file):
    """
    Try to get the date that a file was created, falling back to when it was
    last modified if that isn't possible.
    See http://stackoverflow.com/a/39501288/1709587 for explanation.
    """
    if platform.system() == "Windows":
        return os.path.getctime(path_to_file)
    else:
        stat = os.stat(path_to_file)
        try:
            return stat.st_birthtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return stat.st_mtime



def localtime():
    pass


def mktime():
    pass


def sleep():
    pass


def sleep_ms():
    pass


def sleep_us():
    pass


def ticks_add():
    pass


def ticks_cpu():
    pass


def ticks_diff():
    pass


def ticks_ms():
    pass


def ticks_us():
    pass


def time():
    pass



"""Prints information about how should one run python or pip so that the commands
affect same Python installation that is used for running this script"""

import os.path
import platform
import subprocess
import sys
from shutil import which


def _find_commands(logical_command, reference_output, query_arguments, only_best=True):
    """Returns the commands that can be used for running given conceptual command
    (python or pip)"""

    def is_correct_command(command):
        # Don't try to run the command itself, but first expand it to full path.
        # The location of parent executable seems to affect command search.
        full_path = which(command)
        if full_path is None:
            return False

        try:
            output = subprocess.check_output(
                [full_path] + query_arguments, universal_newlines=True, shell=False
            )

            expected = reference_output.strip()
            actual = output.strip()
            if platform.system() == "Windows":
                expected = expected.lower()
                actual = actual.lower()

            return expected == actual
        except Exception:
            return False

    correct_commands = set()

    # first look for short commands
    for suffix in _get_version_suffixes():
        command_ = logical_command + suffix
        if is_correct_command(command_):
            if " " in command_:
                command_ = '"' + command_ + '"'

            correct_commands.add(command_)
            if only_best:
                return list(correct_commands)

    # if no Python found, then use executable
    if (
        len(correct_commands) == 0
        and logical_command == "python"
        and platform.system() != "Windows"
    ):  # Unixes tend to use symlinks, not Windows
        correct_commands.add(sys.executable)
        if only_best:
            return list(correct_commands)

    # if still nothing found, then add full paths
    if len(correct_commands) == 0:
        if platform.system() == "Windows":
            exe_suffix = ".exe"
        else:
            exe_suffix = ""

        folders = [
            sys.exec_prefix,
            os.path.join(sys.exec_prefix, "bin"),
            os.path.join(sys.exec_prefix, "Scripts"),
        ]

        for suffix in _get_version_suffixes():
            command_ = logical_command + suffix
            for folder in folders:
                full_command = os.path.join(folder, command_)
                if os.path.exists(full_command + exe_suffix):
                    if " " in full_command:
                        full_command = '"' + full_command + '"'

                    correct_commands.add(full_command)
                    if only_best:
                        return list(correct_commands)

    return sorted(correct_commands, key=len)


def _find_python_commands(only_best=True):
    return _find_commands(
        "python",
        sys.exec_prefix + "\n" + sys.version,
        ["-c", "import sys; print(sys.exec_prefix); print(sys.version)"],
        only_best,
    )


def _find_pip_commands(only_best=True):
    # Asking pip version is quite slow.
    # Trying a shortcut for common case:
    #  if $(which <command>) lives in the same dir as current interpreter
    #  and we're using Thonny-private venv,
    #  we can trust the command is the right one.
    pref_cmd = "pip" + _get_version_suffixes()[0]
    pref_cmd_path = which(pref_cmd)
    if pref_cmd_path:
        pref_cmd_dir = os.path.dirname(pref_cmd_path)
        current_exe_dir = os.path.dirname(sys.executable)
        if pref_cmd_dir == current_exe_dir and os.path.isfile(
            os.path.join(current_exe_dir, "is_private")
        ):
            return [pref_cmd]

    # Fallback
    current_ver_string = _get_pip_version_string()

    if current_ver_string is not None:
        commands = _find_commands("pip", current_ver_string, ["--version"], only_best)
        if len(commands) > 0:
            return commands
        else:
            python_commands = _find_python_commands(True)
            return [python_commands[0] + " -m pip"]
    else:
        return []


def _get_version_suffixes():
    major = str(sys.version_info.major)
    minor = "%d.%d" % (sys.version_info.major, sys.version_info.minor)

    if platform.system() == "Windows":
        return ["", major, minor]
    else:
        return [major, minor, ""]


def _get_pip_version_string():
    import io

    try:
        import pip
    except ImportError:
        return None

    # capture output
    original_stdout = sys.stdout
    try:
        sys.stdout = io.StringIO()
        try:
            pip.main(["--version"])
        except SystemExit:
            pass
        return sys.stdout.getvalue().strip()
    finally:
        sys.stdout = original_stdout


def _clear_screen():
    if platform.system() == "Windows":
        os.system("cls")
    else:
        os.system("clear")


if __name__ == "__main__":
    _clear_screen()
    print("*" * 80)
    print(
        "This session is prepared for using Python %s installation in"
        % platform.python_version()
    )
    print(" ", sys.exec_prefix)
    print("")
    print("Command for running the interpreter:")
    for py_command in _find_python_commands(True):
        print(" ", py_command)

    print("")
    print("Command for running pip:")
    # print(_get_pip_version_string())
    pip_commands = _find_pip_commands(True)
    if len(pip_commands) == 0:
        print(" ", "<pip is not installed>")
    else:
        for pip_command in pip_commands:
            print(" ", pip_command)

    print("")
    print("*" * 80)
